"""SOHO Map subclass definitions"""

import numpy as np
from matplotlib.colors import CenteredNorm

import astropy.units as u
from astropy.coordinates import CartesianRepresentation, HeliocentricMeanEcliptic
from astropy.visualization import AsinhStretch, PowerStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy import log
from sunpy.map.mapbase import GenericMap, SpatialPair
from sunpy.map.sources.source_type import source_stretch
from sunpy.time import parse_time

__all__ = ['EITMap', 'EITL1Map', 'LASCOMap', 'MDIMap', 'MDISynopticMap']


class EITMap(GenericMap):
    """
    SOHO EIT Image Map.

    SOHO EIT is an extreme ultraviolet (EUV) imager able to image the solar
    transition region and inner corona in four selected bandpasses,
    171 (Fe IX/X), 195 (Fe XII), 284 (Fe XV), and 304 (He II) Angstrom.

    SOHO was launched on 2 December 2 1995 into a sun-synchronous orbit and
    primary mission operations for SOHO EIT ended at the end of July 2010.

    References
    ----------
    * `SOHO Mission Page <https://sohowww.nascom.nasa.gov/>`__
    * `SOHO EIT Instrument Page <https://umbra.nascom.nasa.gov/eit/>`__
    * `SOHO EIT User Guide <https://umbra.nascom.nasa.gov/eit/eit_guide/>`__

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset

        import sunpy.map
        from sunpy.data.sample import EIT_195_IMAGE
        eit_map = sunpy.map.Map(EIT_195_IMAGE)
        eit_map.plot()
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self._nickname = self.instrument
        self.plot_settings['cmap'] = f"sohoeit{str(int(self.wavelength.to('angstrom').value))}"
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.5)), clip=False)

    @property
    def coordinate_system(self):
        """
        Override the default implementation of coordinate_system to handle EITMAP-specific logic for CTYPE values.
        """
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1.lower() in ("solar-x", "solar_x"):
            ctype1 = 'HPLN-TAN'
        if ctype2.lower() in ("solar-y", "solar_y"):
            ctype2 = 'HPLT-TAN'
        return SpatialPair(ctype1, ctype2)

    @property
    def date(self):
        # Old EIT data has date-obs in format of dd-JAN-yy so we use date_obs where available
        return self._get_date('date_obs') or super().date

    @property
    def spatial_units(self):
        """
        If not present in CUNIT{1,2} keywords, defaults to arcsec.
        """
        return SpatialPair(u.Unit(self.meta.get('cunit1', 'arcsec')),
                           u.Unit(self.meta.get('cunit2', 'arcsec')))

    @property
    def waveunit(self):
        """
        If WAVEUNIT FITS keyword isn't present, defaults to Angstrom.
        """
        unit = self.meta.get("waveunit", "Angstrom") or "Angstrom"
        return u.Unit(unit)

    @property
    def detector(self):
        # Even though this is wrong, we need to keep this for backwards compatibility
        return "EIT"

    @property
    def instrument(self):
        return "EIT"

    @property
    def rsun_obs(self):
        return u.Quantity(self.meta['solar_r'] * self.meta['cdelt1'], 'arcsec')

    @property
    def _supported_observer_coordinates(self):
        return [(('hec_x', 'hec_y', 'hec_z'), {'x': self.meta.get('hec_x'),
                                               'y': self.meta.get('hec_y'),
                                               'z': self.meta.get('hec_z'),
                                               'unit': u.km,
                                               'representation_type': CartesianRepresentation,
                                               'frame': HeliocentricMeanEcliptic})
                ] + super()._supported_observer_coordinates

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an EIT image"""
        return (header.get('instrume') == 'EIT' or header.get('telescop') == 'Extreme-ultraviolet Imaging Telescope (EIT)'  and header.get('level') is None)


class EITL1Map(EITMap):
    """
    SOHO EIT L1 Image Map.

    See the docstring of `EITMap` for information on the EIT instrument.

    The EIT L1 data is a level 1 data product with very different headers to the normal/older EIT data.
    This has required a new class to be created to handle the differences.

    References
    ----------
    * `SOHO Mission Page <https://sohowww.nascom.nasa.gov/>`__
    * `SOHO EIT Instrument Page <https://umbra.nascom.nasa.gov/eit/>`__
    * `SOHO EIT User Guide <https://umbra.nascom.nasa.gov/eit/eit_guide/>`__
    """
    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, AsinhStretch(0.0001)), clip=False)

    @property
    def processing_level(self):
        """
        Returns the FITS processing level if present.
        """
        return self.meta.get('LEVEL', None)

    @property
    def instrument(self):
        """
        Returns the instrument name.
        """
        # EIT L1 has the instrument name in the TELESCOP keyword
        return self.meta.get('TELESCOP')

    @property
    def date(self):
        # EIT L1 data has the date in the DATE-BEG keyword
        # so we need to call out to GenericMap to get the date.
        return super(EITMap, self).date

    @property
    def rsun_obs(self):
        return u.Quantity(self.meta['rsun_obs'], 'arcsec')

    @property
    def _supported_observer_coordinates(self):
        return [(
            ('haex_obs', 'haey_obs', 'haez_obs'),
            {'x': self.meta.get('haex_obs'),
            'y': self.meta.get('haey_obs'),
            'z': self.meta.get('haez_obs'),
            'unit': u.m,
            'representation_type': CartesianRepresentation,
            'frame': HeliocentricMeanEcliptic})
        ] + super()._supported_observer_coordinates

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an EIT L1 Image"""
        return (header.get('instrume') == 'EIT' or header.get('telescop') == 'Extreme-ultraviolet Imaging Telescope (EIT)') and header.get('level') == "L1"


class LASCOMap(GenericMap):
    """
    SOHO LASCO Image Map

    The Large Angle and Spectrometric COronagraph (LASCO) is a set of three
    Lyot-type coronagraphs (C1, C2, and C3) that image the solar corona from
    1.1 to 32 solar radii.

    The C1 images rom 1.1 to 3 solar radii. The C2 telescope images the corona
    from 2 to 6 solar radii, overlapping the outer field-of-view of C1 from 2 to
    3 solar radii. The C3 telescope extends the field-of-view to 32 solar radii.

    SOHO was launched on 2 December 2 1995 into a sun-synchronous orbit.

    References
    ----------
    * `SOHO Mission Page <https://sohowww.nascom.nasa.gov/>`__
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self.plot_settings['cmap'] = f'soholasco{self.detector[1]!s}'
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.5)), clip=False)

    @property
    def coordinate_system(self):
        """
        Override the default implementation to handle LASCOMAP-specific logic for CTYPE values.
        """
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1.lower() in ("solar-x", "solar_x"):
            ctype1 = 'HPLN-TAN'
        if ctype2.lower() in ("solar-y", "solar_y"):
            ctype2 = 'HPLT-TAN'
        return SpatialPair(ctype1, ctype2)

    @property
    def spatial_units(self):
        return SpatialPair(u.Unit(self.meta.get('cunit1').lower()),
                           u.Unit(self.meta.get('cunit2').lower()))

    @property
    def rotation_matrix(self):
        # For Helioviewer images, clear rotation metadata, as these have already been rotated.
        # Also check that all CROTAn keywords exist to make sure that it's an untouched
        # Helioviewer file.
        if ('helioviewer' in self.meta and
                'crota' in self.meta and
                'crota1' in self.meta and
                'crota2' in self.meta):
            log.debug("LASCOMap: Ignoring CROTAn keywords "
                      "because the map has already been rotated by Helioviewer")
            return np.identity(2)
        else:
            return super().rotation_matrix

    @property
    def date(self):
        if date := self.meta.get('date-obs', self.meta.get('date_obs')):
            # If the header has already been fixed, no need to concatenate
            if (time := self.meta.get('time-obs', self.meta.get('time_obs'))) and 'T' not in date:
                date = f"{date}T{time}"
            date = parse_time(date)
        return date or super().date

    def _set_date(self, date):
        if 'time-obs' in self.meta:
            time_key = 'time-obs'
            del self.meta['time-obs']
        if 'time_obs' in self.meta:
            time_key = 'time_obs'
            del self.meta['time_obs']
        date_key = 'date-obs' if 'date-obs' in self.meta else 'date_obs'
        if time_key in self.meta:
            self.meta[date_key], self.meta[time_key] = parse_time(date).utc.isot.split('T')
        else:
            self.meta[date_key] = parse_time(date).utc.isot

    @property
    def nickname(self):
        filter = self.meta.get('filter', '')
        return f'{self.instrument}-{self.detector} {filter}'

    @nickname.setter
    def nickname(self, value):
        raise AttributeError("Cannot manually set nickname for LASCOMap")

    @property
    def measurement(self):
        # TODO: This needs to do more than white-light. Should give B, pB, etc.
        return "white-light"

    @property
    def unit(self):
        bunit = self.meta.get('bunit', None)
        if bunit is not None and bunit == 0:
            # The HV JP2 files given to us have a 0 value BUNIT
            return u.dimensionless_unscaled
        return super().unit

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an LASCO image."""
        return header.get('instrume') == 'LASCO'


class MDIMap(GenericMap):
    """
    SOHO MDI Image Map

    The Michelson Doppler Imager (MDI) is a white light refracting telescope
    which feeds sunlight through a series of filters onto a CCD camera. Two
    tunable Michelson interformeters define a 94 mAngstrom bandpass that can be
    tuned across the Ni 6768 Angstrom solar absorption line.

    MDI measures line-of-sight motion (Dopplergrams), magnetic field
    (magnetograms), and brightness images of the full solar disk at several
    resolutions (4 arc-second to very low resolution) and a fixed selected
    region in higher resolution (1.2 arc-second).

    SOHO was launched on 2 December 2 1995 into a sun-synchronous orbit and
    SOHO MDI ceased normal science observations on 12 April 2011.

    References
    ----------
    * `SOHO Mission Page <https://sohowww.nascom.nasa.gov/>`__
    * `SOHO MDI Instrument Page <http://soi.stanford.edu>`__
    * `SOHO MDI Fits Header keywords <http://soi.stanford.edu/sssc/doc/keywords.html>`__
    * :cite:t:`scherrer_solar_1995`
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        if self.unit is not None and self.unit.is_equivalent(u.T):
            # Magnetic field maps, not intensity maps
            self.plot_settings['norm'] = CenteredNorm()

    @property
    def _date_obs(self):
        if 'T' in self.meta.get('date-obs', ''):
            # Helioviewer MDI files have the full date in DATE_OBS, but we still
            # want to let normal FITS files use DATE-OBS
            return parse_time(self.meta['date-obs'])
        elif 'date_obs' in self.meta:
            return parse_time(self.meta['date_obs'])

    @property
    def unit(self):
        bunit = self.meta.get('bunit', None)
        if bunit is not None and bunit.lower() == 'arbitrary intensity units':
            return u.dimensionless_unscaled
        return super().unit

    @property
    def spatial_units(self):
        """
        If not present in CUNIT{1,2} keywords, defaults to arcsec.
        """
        return SpatialPair(u.Unit(self.meta.get('cunit1', 'arcsec')),
                           u.Unit(self.meta.get('cunit2', 'arcsec')))

    @staticmethod
    def _is_mdi_map(header):
        return header.get('instrume') == 'MDI' or header.get('camera') == 'MDI'

    @staticmethod
    def _is_synoptic_map(header):
        return 'Synoptic Chart' in header.get('CONTENT', '')

    @property
    def _supported_observer_coordinates(self):
        return [(('obs_l0', 'obs_b0', 'obs_dist'), {'lon': self.meta.get('obs_l0'),
                                                    'lat': self.meta.get('obs_b0'),
                                                    'radius': self.meta.get('obs_dist'),
                                                    'unit': (u.deg, u.deg, u.AU),
                                                    'frame': "heliographic_carrington"}),
                ] + super()._supported_observer_coordinates

    @property
    def instrument(self):
        return "MDI"

    @property
    def waveunit(self):
        """
        Always assumed to be Angstrom.
        """
        return "Angstrom"

    @property
    def measurement(self):
        """
        Returns the measurement type.
        """
        return self.meta.get('CONTENT', '')

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an MDI image"""
        return cls._is_mdi_map(header) and not cls._is_synoptic_map(header)


class MDISynopticMap(MDIMap):
    """
    SOHO MDI synoptic magnetogram Map.

    See the docstring of `MDIMap` for information on the MDI instrument.
    """

    @property
    def date(self):
        """
        Image observation time.

        This is taken from the 'DATE-OBS' or 'T_OBS' keywords.
        """
        return self._get_date('date-obs') or self._get_date('t_obs') or super().date

    def _set_date(self, date):
        self.meta['date-obs'] = self.meta['t_obs'] = parse_time(date).utc.isot

    @property
    def spatial_units(self):
        cunit1 = self.meta['cunit1']
        if cunit1 == 'Degree':
            cunit1 = 'deg'

        cunit2 = self.meta['cunit2']
        if cunit2 == 'Sine Latitude':
            cunit2 = 'deg'

        return SpatialPair(u.Unit(cunit1), u.Unit(cunit2))

    @property
    def unit(self):
        bunit = self.meta.get('bunit', None)
        if bunit is None:
            return
        # Maxwells aren't in the IAU unit style manual and therefore not a valid FITS unit
        # The mapbase unit property forces this validation, so we must override it to prevent it.
        return u.Unit(bunit)

    @property
    def scale(self):
        if self.meta['cunit2'] == 'Sine Latitude':
            # Since, this map uses the cylindrical equal-area (CEA) projection,
            # the spacing should be modified to 180/pi times the original value
            # Reference: Section 5.5, Thompson 2006
            return SpatialPair(np.abs(self.meta['cdelt1']) * self.spatial_units[0] / u.pixel,
                               180 / np.pi * self.meta['cdelt2'] * u.deg / u.pixel)

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an MDI image"""
        return cls._is_mdi_map(header) and cls._is_synoptic_map(header)



"""
Source type-dependencies.

Helioviewer JPEG2000 files have stretched images compared to the FITS data.

"""
from astropy.visualization import LinearStretch

__all__ = ['from_helioviewer_project', 'source_stretch']


def from_helioviewer_project(meta):
    """
    Test determining if the given metadata contains Helioviewer Project sourced data.

    Parameters
    ----------
    meta : `~astropy.utils.metadata.MetaData`
        The metadata to parse.

    Returns
    -------
    If the data of the map comes from the Helioviewer Project, then True is
    returned. If not, False is returned.
    """
    return 'helioviewer' in meta.keys()


def source_stretch(meta, fits_stretch):
    """
    Assign the correct source-dependent image stretching function.

    Parameters
    ----------
    meta : `~astropy.utils.metadata.MetaData`
        The metadata to parse.
    fits_stretch : `~astropy.visualization.BaseStretch`
        Image stretching function used when the source image data comes from a
        FITS file.

    Returns
    -------
    An image stretching function appropriate to the image data source.
    """
    if from_helioviewer_project(meta):
        # Helioviewer JPEG2000 files already have a stretched data values, so
        # just use a linear stretch.
        return LinearStretch()
    else:
        # Not a Helioviewer JPEG2000 file, so assume the data has not been
        # stretched and so use the FITS stretching as defined in the instrument
        # source.
        return fits_stretch



"""
===========================================================
Identifying stars in a STEREO/SECCHI COR2 coronagraph image
===========================================================

Since the field of view in SECCHI COR2 images can span 2 to 15 solar radii,
we often observe stars in these image data. In this example, we will use the
`Astroquery package <https://astroquery.readthedocs.io/en/latest/>`__ to query the
`VizieR star catalog <https://vizier.cds.unistra.fr/viz-bin/VizieR>`__ for stars observed
by the `Gaia satellite <https://www.esa.int/Science_Exploration/Space_Science/Gaia>`__ within
the SECCHI COR2 field of view. Then we will use the coordinates framework in SunPy and AstroPy
to transform the coordinates returned by VizieR into SECCHI COR2 image coordinates.
As a bonus, we'll also identify Mars."

This requires the installation of the `astroquery <https://astroquery.readthedocs.io/en/latest/>`__
package, which can be installed on top of the existing sunpy conda
environment: ``conda install -c astropy astroquery`` and an active internet connection.
"""
import hvpy
import matplotlib.pyplot as plt
from astroquery.gaia import Gaia

import astropy.units as u
from astropy.coordinates import Distance, Longitude, SkyCoord
from astropy.time import Time

import sunpy.map
from sunpy.coordinates import get_body_heliographic_stonyhurst
from sunpy.time import parse_time
from sunpy.util.config import get_and_create_download_dir

###############################################################################
# Let's download a STEREO-A SECCHI COR2 image from Helioviewer.org which provide
# pre-processed images and load it into a Map.
# We download to the default sunpy download directory.

cor2_file = hvpy.save_file(hvpy.getJP2Image(parse_time('2014/05/15 07:54').datetime,
                                            hvpy.DataSource.COR2_A.value),
                           get_and_create_download_dir() + "/COR2.jp2")
cor2_map = sunpy.map.Map(cor2_file)

###############################################################################
# To efficiently search the star field, we need to know what stars are near the
# Sun as observed by STEREO. We need the vector that points from STEREO to the Sun.
# The location of STEREO in HCRS provides the Sun-to-STEREO vector.

sun_to_stereo = cor2_map.observer_coordinate.transform_to('hcrs')

###############################################################################
# We then calculate the apparent right ascension and declination of the Sun as
# seen from STEREO.

center_ra = Longitude(sun_to_stereo.ra + 180*u.deg).to_value(u.deg)
center_dec = -sun_to_stereo.dec.to_value(u.deg)
print(center_ra, center_dec)

###############################################################################
# Let's look up bright stars using the Gaia search capability provided by
# `astroquery <https://astroquery.readthedocs.io/en/latest/>`__.
# We will search the Gaia DR3 star catalogue for stars with magnitude
# brighter than 7.5 within 4 degrees of the Sun direction.

job = Gaia.launch_job(
    "SELECT ra, dec, phot_g_mean_mag, parallax, pmra, pmdec, radial_velocity, ref_epoch "
    "FROM gaiadr3.gaia_source "
    "WHERE phot_g_mean_mag < 7.5 "
    f"AND CONTAINS(POINT('ICRS', ra, dec), CIRCLE('ICRS', {center_ra}, {center_dec}, {4})) = 1"
)
result = job.get_results()

###############################################################################
# Let's see how many stars we've found.

print(len(result))

###############################################################################
# Now we load all stars into an array coordinate. The reference epoch for the
# star positions is J2016.0, so we update these positions to the date of the
# COR2 observation using :meth:`astropy.coordinates.SkyCoord.apply_space_motion`.
# Some Gaia entries may be missing information on parallax, proper motion, or
# radial velocity, so we fill those gaps with the fallback assumptions of
# negligible parallax and zero velocity.

tbl_crds = SkyCoord(ra=result['ra'],
                    dec=result['dec'],
                    distance=Distance(parallax=u.Quantity(result['parallax'].filled(1e-3))),
                    pm_ra_cosdec=result['pmra'].filled(0),
                    pm_dec=result['pmdec'].filled(0),
                    radial_velocity=result['radial_velocity'].filled(0),
                    frame='icrs',
                    obstime=Time(result['ref_epoch'], format='jyear'))
tbl_crds = tbl_crds.apply_space_motion(new_obstime=cor2_map.date)

###############################################################################
# One of the bright features is actually Mars, so let's also get that coordinate.

mars = get_body_heliographic_stonyhurst('mars', cor2_map.date, observer=cor2_map.observer_coordinate)

###############################################################################
# Let's plot the results. The coordinates will be transformed automatically
# when plotted using :meth:`~astropy.visualization.wcsaxes.WCSAxes.plot_coord`.

fig = plt.figure()
ax = fig.add_subplot(projection=cor2_map)

# Let's tweak the axis to show in degrees instead of arcsec
lon, lat = ax.coords
lon.set_major_formatter('d.dd')
lat.set_major_formatter('d.dd')

cor2_map.plot(axes=ax, vmin=0, vmax=600)
cor2_map.draw_limb(axes=ax)

# Plot the position of Mars
ax.plot_coord(mars, 's', color='white', fillstyle='none', markersize=12, label='Mars')
# Plot all of the stars
ax.plot_coord(tbl_crds, 'o', color='white', fillstyle='none')
ax.legend()

plt.show()



"""
Test cases for HINODE XRTMap subclass.
"""
import copy

import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.mapbase import SpatialPair
from sunpy.map.sources.hinode import XRTMap
from .helpers import _test_private_date_setters


@pytest.fixture
def xrt_map():
    return get_dummy_map_from_header(get_test_filepath("HinodeXRT.header"))


def test_fits_to_xrt(xrt_map):
    """Tests the creation of XRTMap using FITS."""
    assert isinstance(xrt_map, XRTMap)


def test_xrt_coordinate_system(xrt_map):
    assert xrt_map.coordinate_system ==  SpatialPair(axis1='HPLN-TAN', axis2='HPLT-TAN')


def test_is_datasource_for(xrt_map):
    """Test the is_datasource_for method of XRTMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert xrt_map.is_datasource_for(xrt_map.data, xrt_map.meta)


def test_reference_date(xrt_map):
    assert xrt_map.reference_date.isot == "2006-11-11T00:00:19.141"


def test_date(xrt_map):
    assert xrt_map.date.isot == "2006-11-11T00:00:19.141"


def test_private_date_setters(xrt_map):
    _test_private_date_setters(xrt_map)


def test_observatory(xrt_map):
    """Tests the observatory property of the XRTMap object."""
    assert xrt_map.observatory == "Hinode"


def test_measurement(xrt_map):
    """Tests the measurement property of the XRTMap object."""
    assert xrt_map.measurement == 'Be thin-Open'


def test_unit(xrt_map):
    """Tests the unit property of the XRTMap object."""
    assert xrt_map.unit == u.DN / u.second

def test_unit_no_renormalize(xrt_map):
    """
    Tests that the unit defaults ot DN if the history key does not say it has
    been normalized
    """
    new_xrt_map = copy.deepcopy(xrt_map)
    new_xrt_map.meta.pop('history')
    assert new_xrt_map.unit == u.DN

def test_level_number(xrt_map):
    assert xrt_map.processing_level == 1


def test_heliographic_longitude(xrt_map):
    assert u.allclose(xrt_map.heliographic_longitude, 0 * u.deg)


def test_heliographic_latitude(xrt_map):
    assert u.allclose(xrt_map.heliographic_latitude, 3.33047459 * u.deg)


def test_wheel_measurements(xrt_map):
    """Tests the filter_wheel_measurements objects present
    in the XRTMap object."""
    assert (xrt_map.filter_wheel1_measurements ==
            ["Al_med", "Al_poly", "Be_med", "Be_thin", "C_poly", "Open"])
    assert (xrt_map.filter_wheel2_measurements ==
            ["Open", "Al_mesh", "Al_thick", "Be_thick", "Gband", "Ti_poly"])


def test_wcs(xrt_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    xrt_map.pixel_to_world(0*u.pix, 0*u.pix)



import datetime
from collections import OrderedDict

import numpy as np
import pytest
from erfa.core import ErfaWarning
from pandas import DataFrame

import astropy.units as u
from astropy.table import Table
from astropy.time import TimeDelta

import sunpy
import sunpy.timeseries
from sunpy.data.test import get_test_data_filenames, get_test_filepath
from sunpy.time import parse_time
from sunpy.util import SunpyUserWarning
from sunpy.util.metadata import MetaDict

# =============================================================================
# TimeSeries Tests
# =============================================================================

eve_filepath = get_test_filepath('EVE_L0CS_DIODES_1m_truncated.txt')
esp_filepath = get_test_filepath('eve_l1_esp_2011046_00_truncated.fits')
fermi_gbm_filepath = get_test_filepath('gbm.fits')
norh_filepath = get_test_filepath('tca110810_truncated')
lyra_filepath = get_test_filepath('lyra_20150101-000000_lev3_std_truncated.fits.gz')
rhessi_filepath = get_test_filepath('hsi_obssumm_20120601_018_truncated.fits.gz')
noaa_ind_json_filepath = get_test_filepath('observed-solar-cycle-indices-truncated.json')
noaa_pre_json_filepath = get_test_filepath('predicted-solar-cycle-truncated.json')
goes_filepath = get_test_filepath('go1520120601.fits.gz')
a_list_of_many = [f for f in get_test_data_filenames() if f.parents[0].relative_to(f.parents[1]).name == 'eve']


@pytest.fixture
def eve_test_ts():
    with pytest.warns(SunpyUserWarning, match='Unknown units'):
        return sunpy.timeseries.TimeSeries(eve_filepath, source='EVE')


@pytest.fixture
def esp_test_ts():
    return sunpy.timeseries.TimeSeries(esp_filepath, source='ESP')


@pytest.fixture
def fermi_gbm_test_ts():
    return sunpy.timeseries.TimeSeries(fermi_gbm_filepath, source='GBMSummary')


@pytest.fixture
def norh_test_ts():
    return sunpy.timeseries.TimeSeries(norh_filepath, source='NoRH')


@pytest.fixture
def goes_test_ts():
    return sunpy.timeseries.TimeSeries(goes_filepath, source='XRS')


@pytest.fixture
def lyra_test_ts():
    return sunpy.timeseries.TimeSeries(lyra_filepath, source='LYRA')


@pytest.fixture
def rhessi_test_ts():
    return sunpy.timeseries.TimeSeries(rhessi_filepath, source='RHESSI')


@pytest.fixture
def noaa_ind_json_test_ts():
    return sunpy.timeseries.TimeSeries(noaa_ind_json_filepath, source='NOAAIndices')


@pytest.fixture
def noaa_pre_json_test_ts():
    # NOAA pre data contains years long into the future, which ERFA complains about
    with pytest.warns(ErfaWarning, match=r'.*dubious year'):
        return sunpy.timeseries.TimeSeries(
            noaa_pre_json_filepath, source='NOAAPredictIndices')


@pytest.fixture
def generic_ts():
    # Generate the data and the corresponding dates
    base = parse_time("2016/10/01T05:00:00")
    dates = base - TimeDelta(np.arange(24 * 60)*u.minute)
    intensity = np.sin(np.arange(0, 12 * np.pi, ((12 * np.pi) / (24 * 60))))
    intensity2 = np.cos(np.arange(0, 12 * np.pi, ((12 * np.pi) / (24 * 60))))

    # Create the data DataFrame, header MetaDict and units dict
    data = DataFrame(np.column_stack([intensity, intensity2]),
                     index=dates.isot.astype('datetime64'),
                     columns=['intensity', 'intensity2'])
    units = {'intensity': u.W / u.m**2,
             'intensity2': u.W / u.m**2}
    meta = MetaDict({'key': 'value'})

    # Create the time series
    return sunpy.timeseries.TimeSeries(data, meta, units)


@pytest.fixture
def concatenate_multi_files_ts():
    with pytest.warns(SunpyUserWarning, match='Unknown units'):
        return sunpy.timeseries.TimeSeries(
            a_list_of_many, source='EVE', concatenate=True)

# =============================================================================
# Test Creating TimeSeries From Various Dataformats
# =============================================================================


@pytest.fixture
def table_ts():
    # Generate the data and the corresponding dates
    base = parse_time(datetime.datetime.today())
    times = base - TimeDelta(np.arange(24 * 60)*u.minute)
    intensity = u.Quantity(
        np.sin(np.arange(0, 12 * np.pi, ((12 * np.pi) / (24 * 60)))), u.W / u.m ** 2)

    # Create the units and meta objects
    units = OrderedDict([('intensity', u.W / u.m**2)])
    meta = MetaDict({'key': 'value'})
    tbl_meta = MetaDict({'t_key': 't_value'})

    # Create a suitable mixin qtable
    table = Table(
        [times, intensity], names=['time', 'intensity'], meta=tbl_meta)
    table.add_index('time')

    # Create TS from dataframe and check
    return sunpy.timeseries.TimeSeries(table, meta, units)


# =============================================================================
# Test Resulting TimeSeries Parameters
# =============================================================================
@pytest.fixture(params=['eve_test_ts', 'esp_test_ts', 'fermi_gbm_test_ts', 'norh_test_ts', 'goes_test_ts',
                        'lyra_test_ts', 'rhessi_test_ts', 'noaa_ind_json_test_ts',
                        'noaa_pre_json_test_ts', 'generic_ts', 'table_ts'])
def many_ts(request):
    # Fixture to return lots of different timeseries
    return request.getfixturevalue(request.param)



import sys
import platform
import warnings
from pathlib import Path
from functools import wraps
from importlib.metadata import entry_points

import matplotlib as mpl
import matplotlib.pyplot as plt
import pytest

import astropy
from astropy.wcs.wcs import FITSFixedWarning

# NOTE: Do not import sunpy subpackages which have optional dependencies here,
# this module should be importable with no extra dependencies installed.

__all__ = ['skip_windows', 'skip_glymur', 'skip_ana', 'skip_cdf', 'skip_opencv', 'warnings_as_errors', 'asdf_entry_points', 'skip_jsoc']

try:
    import glymur
except ImportError:
    SKIP_GLYMUR = True
else:
    # See if we have a C backend installed.
    # Glymur will not be able to read JPEG2000 files without it.
    SKIP_GLYMUR = not glymur.lib.openjp2.OPENJP2

try:
    from sunpy.io import _pyana  # NOQA
    SKIP_ANA = False
except ImportError:
    SKIP_ANA = True

if sys.maxsize > 2**32:
    SKIP_32 = False
else:
    SKIP_32 = True

try:
    import cv2  # NOQA
    SKIP_OPENCV = False
except ImportError:
    SKIP_OPENCV = True

try:
    import skimage  # NOQA
    SKIP_SKIMAGE = False
except ImportError:
    SKIP_SKIMAGE = True

try:
    import cdflib  # NOQA
    SKIP_CDF = False
except ImportError:
    SKIP_CDF = True

try:
    import asdf  # NOQA
    SKIP_ASDF = False
except ImportError:
    SKIP_ASDF = True


asdf_entry_points = pytest.mark.skipif(
    SKIP_ASDF or not entry_points().select(group="asdf.resource_mappings", name="sunpy"),
    reason="No SunPy ASDF entry points.",
)
skip_ana = pytest.mark.skipif(SKIP_ANA, reason="ANA is not available (or SUNPY_NO_BUILD_ANA_EXTENSION is set).")
skip_cdf = pytest.mark.skipif(SKIP_CDF, reason="CDFlib is not available.")
skip_glymur = pytest.mark.skipif(SKIP_GLYMUR, reason="Glymur can not be imported.")
skip_jsoc = pytest.mark.skip(reason="JSOC is not available.")
skip_opencv = pytest.mark.skipif(SKIP_OPENCV, reason="opencv is not available.")
skip_skimage = pytest.mark.skipif(SKIP_SKIMAGE, reason="skimage is not available.")
skip_windows = pytest.mark.skipif(platform.system() == "Windows", reason="Windows.")



@pytest.fixture
def warnings_as_errors():
    warnings.simplefilter('error')
    yield
    warnings.resetwarnings()


def get_hash_library_name():
    """
    Generate the hash library name for this env.
    """
    import mpl_animators

    animators_version = "dev" if (("dev" in mpl_animators.__version__) or ("rc" in mpl_animators.__version__)) else mpl_animators.__version__.replace('.', '')
    ft2_version = f"{mpl.ft2font.__freetype_version__.replace('.', '')}"
    mpl_version = "dev" if (("dev" in mpl.__version__) or ("rc" in mpl.__version__)) else mpl.__version__.replace('.', '')
    astropy_version = "dev" if (("dev" in astropy.__version__) or ("rc" in astropy.__version__)) else astropy.__version__.replace('.', '')
    return f"figure_hashes_mpl_{mpl_version}_ft_{ft2_version}_astropy_{astropy_version}_animators_{animators_version}.json"


def figure_test(test_function):
    """
    A decorator for a test that verifies the hash of the current figure or the
    returned figure, with the name of the test function as the hash identifier
    in the library. A PNG is also created in the 'result_image' directory,
    which is created on the current path.

    All such decorated tests are marked with `pytest.mark.mpl_image` for convenient filtering.

    Examples
    --------
    @figure_test
    def test_simple_plot():
        plt.plot([0,1])
    """
    hash_library_name = get_hash_library_name()
    hash_library_file = Path(__file__).parent / hash_library_name

    @pytest.mark.mpl_image_compare(hash_library=hash_library_file,
                                   savefig_kwargs={'metadata': {'Software': None}},
                                   style='default')
    @wraps(test_function)
    def test_wrapper(*args, **kwargs):
        ret = test_function(*args, **kwargs)
        if ret is None:
            ret = plt.gcf()
        return ret
    return test_wrapper


def no_vso(f):
    """
    Disable the VSO client from returning results via Fido during this test.
    """
    from sunpy.net import Fido
    from sunpy.net.vso import VSOClient

    @wraps(f)
    def wrapper(*args, **kwargs):
        Fido.registry[VSOClient] = lambda *args: False
        res = f(*args, **kwargs)
        Fido.registry[VSOClient] = VSOClient._can_handle_query
        return res
    return wrapper


def fix_map_wcs(smap):
    import sunpy.map

    # Helper function to fix a WCS and silence the warnings
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=FITSFixedWarning)
        wcs = smap.wcs
        wcs.fix()
    return sunpy.map.Map(smap.data, wcs)



import os
import pathlib
import tempfile

import matplotlib.pyplot as plt
import numpy as np
import pytest

import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS

import sunpy
import sunpy.map
from sunpy.data.test import get_dummy_map_from_header, get_test_data_filenames, get_test_filepath, rootdir
from sunpy.io._fits import read
from sunpy.tests.helpers import asdf_entry_points, figure_test, skip_glymur
from sunpy.util.exceptions import NoMapsInFileError, SunpyMetadataWarning, SunpyUserWarning

a_list_of_many = [f for f in get_test_data_filenames() if 'efz' in f.name]

AIA_171_IMAGE = get_test_filepath('aia_171_level1.fits')
RHESSI_IMAGE = get_test_filepath('hsi_image_20101016_191218.fits')
AIA_193_JP2 = get_test_filepath("2013_06_24__17_31_30_84__SDO_AIA_AIA_193.jp2")
HMI_LOS_JP2 = get_test_filepath("2023_01_31__03_39_23_200__SDO_HMI_HMI_continuum.jp2")
AIA_ASDF = get_test_filepath("aiamap_genericmap_1.0.0.asdf")
AIA_MAP = sunpy.map.Map(AIA_171_IMAGE)
VALID_MAP_INPUTS = [
    (AIA_171_IMAGE, ),
    (pathlib.Path(AIA_171_IMAGE), ),
    (rootdir / "EIT", ),
    (os.fspath(rootdir / "EIT"), ),
    (rootdir / "EIT" / "*.fits", ),
    (AIA_MAP, ),
    (AIA_MAP.data, AIA_MAP.meta),
    ((AIA_MAP.data, AIA_MAP.meta), ),
]


@pytest.mark.parametrize('args1', VALID_MAP_INPUTS)
@pytest.mark.parametrize('args2', VALID_MAP_INPUTS)
def test_two_map_inputs(args1, args2):
    out = sunpy.map.Map(*args1, *args2)
    if isinstance(out, list):
        for m in out:
            assert isinstance(m, sunpy.map.GenericMap)
    else:
        assert isinstance(out, sunpy.map.GenericMap)


@asdf_entry_points
def test_read_asdf_and_verify(tmpdir):
    loaded_asdf_map = sunpy.map.Map(AIA_ASDF)
    assert isinstance(loaded_asdf_map.data, np.ndarray)
    assert isinstance(loaded_asdf_map.meta, dict)
    assert isinstance(loaded_asdf_map, sunpy.map.sources.AIAMap)


@asdf_entry_points
def test_map_meta_changes_in_asdf(tmpdir):
    map = sunpy.map.Map(AIA_171_IMAGE)
    map = map.rotate(90 * u.deg)

    assert "pc1_2" in map.meta.added_items
    assert "crota2" in map.meta.removed_items
    assert "crval1" in map.meta.modified_items

    map.save(f"{tmpdir}/check.asdf")
    map_in_asdf = sunpy.map.Map(f"{tmpdir}/check.asdf")

    assert dict(map_in_asdf.meta) == dict(map.meta)


def test_mapsequence(eit_fits_directory):
    # Test making a MapSequence
    sequence = sunpy.map.Map(list(eit_fits_directory.glob('*.fits')), sequence=True)
    assert isinstance(sequence, sunpy.map.MapSequence)


def test_mapsequence_sortby(eit_fits_directory):
    # Test making a MapSequence with sortby kwarg
    sequence = sunpy.map.Map(list(eit_fits_directory.glob('*.fits')), sequence=True, sortby=None)
    assert isinstance(sequence, sunpy.map.MapSequence)


def test_composite():
    # Test making a CompositeMap
    comp = sunpy.map.Map(AIA_171_IMAGE, RHESSI_IMAGE, composite=True)
    assert isinstance(comp, sunpy.map.CompositeMap)

# Want to check that patterns work, so ignore this warning that comes from
# the AIA test data


@pytest.mark.filterwarnings("ignore:Invalid 'BLANK' keyword in header")
def test_patterns(eit_fits_directory):
    # Test different Map pattern matching

    # File name
    aiamap = sunpy.map.Map(AIA_171_IMAGE)
    assert isinstance(aiamap, sunpy.map.GenericMap)

    # Directory
    maps = sunpy.map.Map(os.fspath(eit_fits_directory))
    assert isinstance(maps, list)
    assert ([isinstance(amap, sunpy.map.GenericMap) for amap in maps])

    # Test that returned maps are sorted
    files_sorted = sorted(list(eit_fits_directory.glob('*')))
    maps_sorted = [sunpy.map.Map(os.fspath(f)) for f in files_sorted]
    assert all(m.date == m_s.date for m, m_s in zip(maps, maps_sorted))

    # Pathlib
    path = pathlib.Path(AIA_171_IMAGE)
    aiamap = sunpy.map.Map(path)

    # Generator
    maps = sunpy.map.Map(eit_fits_directory.glob('*.fits'))
    assert isinstance(maps, list)
    assert isinstance(aiamap, sunpy.map.GenericMap)

    # A directory
    maps = sunpy.map.Map(eit_fits_directory)
    assert isinstance(maps, list)
    assert ([isinstance(amap, sunpy.map.GenericMap) for amap in maps])

    # Glob
    pattern = os.path.join(eit_fits_directory, "*")
    maps = sunpy.map.Map(pattern)
    assert isinstance(maps, list)
    assert ([isinstance(amap, sunpy.map.GenericMap) for amap in maps])

    # Test that returned maps are sorted
    # Sorting based on Strings rather than Path Objects which makes the sorting os independent
    # This was added because on Windows the sorting was different than on Linux
    # Windows(and possibly other OSs) use a case-insensitive sort, while Linux uses a case-sensitive sort
    files_sorted = sorted(str(file) for file in pathlib.Path(pattern).parent.glob('*'))
    maps_sorted = [sunpy.map.Map(os.fspath(f)) for f in files_sorted]
    assert all(m.date == m_s.date for m, m_s in zip(maps, maps_sorted))

    # Single character wildcard (?)
    pattern = os.path.join(eit_fits_directory, "efz20040301.0?0010_s.fits")
    maps = sunpy.map.Map(pattern)
    assert isinstance(maps, list)
    assert len(maps) == 7
    assert ([isinstance(amap, sunpy.map.GenericMap) for amap in maps])

    # Character ranges
    pattern = os.path.join(eit_fits_directory, "efz20040301.0[2-6]0010_s.fits")
    maps = sunpy.map.Map(pattern)
    assert isinstance(maps, list)
    assert len(maps) == 4
    assert ([isinstance(amap, sunpy.map.GenericMap) for amap in maps])

    # Already a Map
    amap = sunpy.map.Map(maps[0])
    assert isinstance(amap, sunpy.map.GenericMap)

    # A list of filenames
    maps = sunpy.map.Map(list(eit_fits_directory.glob('*.fits')))
    assert isinstance(maps, list)
    assert ([isinstance(amap, sunpy.map.GenericMap) for amap in maps])

    # Data-header pair in a tuple
    pair_map = sunpy.map.Map((amap.data, amap.meta))
    assert isinstance(pair_map, sunpy.map.GenericMap)

    # Data-header pair not in a tuple
    pair_map = sunpy.map.Map(amap.data, amap.meta)
    assert isinstance(pair_map, sunpy.map.GenericMap)

    # Data-wcs object pair in tuple
    pair_map = sunpy.map.Map((amap.data, WCS(AIA_171_IMAGE, fix=False)))
    assert isinstance(pair_map, sunpy.map.GenericMap)

    # Data-wcs object pair not in a tuple
    pair_map = sunpy.map.Map(amap.data, WCS(AIA_171_IMAGE, fix=False))
    assert isinstance(pair_map, sunpy.map.GenericMap)

    # Data-header from FITS
    with fits.open(AIA_171_IMAGE) as hdul:
        data = hdul[0].data
        header = hdul[0].header
    pair_map = sunpy.map.Map((data, header))
    assert isinstance(pair_map, sunpy.map.GenericMap)
    pair_map, pair_map = sunpy.map.Map(((data, header), (data, header)))
    assert isinstance(pair_map, sunpy.map.GenericMap)
    pair_map = sunpy.map.Map(data, header)
    assert isinstance(pair_map, sunpy.map.GenericMap)

    # Custom Map
    data = np.arange(0, 100).reshape(10, 10)
    header = {'cdelt1': 10, 'cdelt2': 10,
              'telescop': 'sunpy',
              'cunit1': 'arcsec', 'cunit2': 'arcsec'}
    with pytest.warns(SunpyMetadataWarning, match='Missing CTYPE'):
        pair_map = sunpy.map.Map(data, header)
    assert isinstance(pair_map, sunpy.map.GenericMap)

    # Common keys not strings
    data = np.arange(0, 100).reshape(10, 10)
    header = {'cdelt1': 10, 'cdelt2': 10,
              'telescop': 100,
              'detector': 1,
              'instrume': 50,
              'cunit1': 'arcsec', 'cunit2': 'arcsec'}
    with pytest.warns(SunpyMetadataWarning, match='Missing CTYPE'):
        pair_map = sunpy.map.Map(data, header)
    assert isinstance(pair_map, sunpy.map.GenericMap)


def test_errors(tmpdir):
    # If directory doesn't exist, make sure it's listed in the error msg
    nonexist_dir = 'nonexist'
    directory = pathlib.Path(tmpdir, nonexist_dir)
    with pytest.raises(ValueError, match=nonexist_dir):
        sunpy.map.Map(os.fspath(directory))

    with pytest.raises(ValueError, match='Invalid input: 78'):
        # Check a random unsupported type (int) fails
        sunpy.map.Map(78)

    # If one file failed to load, make sure it's raised as an exception.
    p = tmpdir.mkdir("sub").join("hello.fits")
    p.write("content")
    files = [AIA_171_IMAGE, p.strpath]
    with pytest.raises(OSError, match=(r"Failed to read *")):
        sunpy.map.Map(files)


@pytest.mark.filterwarnings("ignore:One of the data, header pairs failed to validate")
@pytest.mark.parametrize(('allow_errors', 'error', 'match'),
                         [(True, RuntimeError, 'No maps loaded'),
                          (False, sunpy.map.mapbase.MapMetaValidationError,
                           'Image coordinate units for axis 1 not present in metadata.')])
def test_allow_errors(allow_errors, error, match):
    # Check that the correct errors are raised depending on allow_errors value
    data = np.arange(0, 100).reshape(10, 10)
    with pytest.raises(error, match=match):
        sunpy.map.Map(data, {}, allow_errors=allow_errors)


def test_dask_array():
    dask_array = pytest.importorskip('dask.array')
    amap = sunpy.map.Map(AIA_171_IMAGE)
    da = dask_array.from_array(amap.data, chunks=(1, 1))
    pair_map = sunpy.map.Map(da, amap.meta)
    assert isinstance(pair_map, sunpy.map.GenericMap)


@pytest.mark.remote_data
def test_url_pattern():
    # A URL
    amap = sunpy.map.Map("http://data.sunpy.org/sample-data/AIA20110319_105400_0171.fits")
    assert isinstance(amap, sunpy.map.GenericMap)


@pytest.mark.remote_data
def test_uri_pattern():
    """
    Testing publicly accessible s3 object
    """
    amap = sunpy.map.Map("s3://data.sunpy.org/sunpy/AIA20110607_065843_0193_cutout.fits", fsspec_kwargs={"anon": True})
    assert isinstance(amap, sunpy.map.GenericMap)


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore:datetime.datetime.utcnow:DeprecationWarning")
@pytest.mark.filterwarnings("ignore:The ANA reader may be removed in a future version of sunpy")
def test_uri_directory_pattern():
    """
    Testing publicly accessible s3 directory
    """
    with pytest.warns(SunpyUserWarning, match='Failed to read'):
        amap = sunpy.map.Map('s3://data.sunpy.org/aiapy', fsspec_kwargs={'anon':True}, allow_errors=True)
    assert all(isinstance(am, sunpy.map.GenericMap) for am in amap)


def test_save():
    # Test save out
    aiamap = sunpy.map.Map(AIA_171_IMAGE)
    afilename = tempfile.NamedTemporaryFile(suffix='fits').name
    with pytest.warns(fits.verify.VerifyWarning, match="Invalid 'BLANK' keyword in header."):
        aiamap.save(afilename, filetype='fits', overwrite=True)
    backin = sunpy.map.Map(afilename)
    assert isinstance(backin, sunpy.map.sources.AIAMap)


@pytest.mark.remote_data
def test_map_list_urls_cache():
    """
    Test for https://github.com/sunpy/sunpy/issues/4006
    """
    urls = ['https://github.com/sunpy/data/raw/main/sunpy/v1/AIA20110607_063305_0094_lowres.fits',
            'https://github.com/sunpy/data/raw/main/sunpy/v1/AIA20110607_063305_0094_lowres.fits']
    with pytest.warns(fits.verify.VerifyWarning, match="Invalid 'BLANK' keyword in header."):
        sunpy.map.Map(urls)


@pytest.mark.remote_data
def test_map_list_uri():
    """
    Test for reading from URI (AWS S3).
    """
    uri_list = ["s3://data.sunpy.org/aiapy/aia_lev1_171a_2019_01_01t00_00_09_35z_image_lev1.fits",
                "s3://data.sunpy.org/aiapy/aia_lev1_94a_2019_01_01t00_00_11_12z_image_lev1.fits"]
    amap = sunpy.map.Map(uri_list, fsspec_kwargs={'anon':True})
    assert all(isinstance(am, sunpy.map.GenericMap) for am in amap)


@pytest.mark.filterwarnings('ignore:File may have been truncated')
@pytest.mark.parametrize(('file', 'mapcls'), [
    ("EIT_header/efz20040301.000010_s.header", sunpy.map.sources.EITMap),
    ("lasco_c2_25299383_s.header", sunpy.map.sources.LASCOMap),
    ("mdi.fd_Ic.20101015_230100_TAI.data.header", sunpy.map.sources.MDIMap),
    ("mdi.fd_M_96m_lev182.20101015_191200_TAI.data.header", sunpy.map.sources.MDIMap),
    ("euvi_20090615_000900_n4euA_s.header", sunpy.map.sources.EUVIMap),
    ("cor1_20090615_000500_s4c1A.header", sunpy.map.sources.CORMap),
    ("hi_20110910_114721_s7h2A.header", sunpy.map.sources.HIMap),
    (AIA_171_IMAGE, sunpy.map.sources.AIAMap),
    (RHESSI_IMAGE, sunpy.map.sources.RHESSIMap),
    ("FGMG4_20110214_030443.7.header", sunpy.map.sources.SOTMap),
    ("swap_lv1_20140606_000113.header", sunpy.map.sources.SWAPMap),
    ("HinodeXRT.header", sunpy.map.sources.XRTMap),
])
def test_sources(file, mapcls):
    p = pathlib.Path(get_test_filepath(file))
    m = get_dummy_map_from_header(p) if p.suffix == '.header' else sunpy.map.Map(p)
    assert isinstance(m, mapcls)


@pytest.mark.skipif(pytest.__version__ < "8.0.0", reason="pytest >= 8.0.0 raises a warning for this test")
def test_no_2d_hdus(tmpdir):
    # Create a fake FITS file with a valid header but 1D data
    tmp_fpath = str(tmpdir / 'data.fits')
    with fits.open(AIA_171_IMAGE, ignore_blank=True) as hdul:
        fits.writeto(tmp_fpath, np.arange(100), hdul[0].header)

    with pytest.raises(NoMapsInFileError, match='Found no HDUs with >= 2D data'):
        sunpy.map.Map(tmp_fpath)


@skip_glymur
def test_map_jp2_AIA():
    jp2_map = sunpy.map.Map(AIA_193_JP2, memmap=False)
    assert isinstance(jp2_map, sunpy.map.GenericMap)
    # The base of an array that owns its memory is None
    # For some reason JP2 doesn't own its own memory but are not type of memmaps.
    assert jp2_map.data.base is not None
    jp2_map = sunpy.map.Map(AIA_193_JP2, memmap=True)
    assert isinstance(jp2_map, sunpy.map.GenericMap)
    assert jp2_map.data.base is not None


@skip_glymur
@figure_test
def test_map_jp2_HMI():
    # We failed to read the HMI JP2 file with map before
    # See https://github.com/sunpy/sunpy/issues/6709
    jp2_map = sunpy.map.Map(HMI_LOS_JP2)
    assert isinstance(jp2_map, sunpy.map.GenericMap)
    jp2_map.plot()
    plt.colorbar()


def test_map_fits(mocker):

    hdus = read(AIA_171_IMAGE)
    mock = mocker.patch("sunpy.io._file_tools.fits.read", return_value=hdus)
    sunpy.map.Map(AIA_171_IMAGE, memmap=False)
    assert mock.call_args.kwargs["memmap"] is False

    mock.reset_mock()
    sunpy.map.Map(AIA_171_IMAGE, memmap=True)
    assert mock.call_args.kwargs["memmap"] is True


def test_map_list_of_files_with_one_broken():
    files = [AIA_171_IMAGE, get_test_filepath('not_actually_fits.fits')]
    with pytest.warns(SunpyUserWarning, match='Failed to read'):
        amap = sunpy.map.Map(files, allow_errors=True)
    assert amap.data.shape == (128, 128)

    files = [AIA_171_IMAGE, get_test_filepath('not_actually_fits.fits'), AIA_171_IMAGE]
    with pytest.warns(SunpyUserWarning, match='Failed to read'):
        amap = sunpy.map.Map(files, allow_errors=True)
    assert len(amap) == 2

    with pytest.warns(SunpyUserWarning, match='Failed to read'):
        amap = sunpy.map.Map(files, allow_errors=True, sequence=True)
    assert len(amap) == 2

    with pytest.raises(OSError, match='Failed to read'):
        sunpy.map.Map(files, allow_errors=False)



import logging
import datetime
from urllib.error import URLError, HTTPError
from unittest.mock import Mock, patch

import pytest

from sunpy.data.test import rootdir
from sunpy.extern import parse
from sunpy.net.scraper import Scraper
from sunpy.net.scraper_utils import get_timerange_from_exdict
from sunpy.time import TimeRange, parse_time


def test_directory_date_pattern():
    s = Scraper(format='{{year:4d}}/{{month:2d}}/{{day:2d}}/{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}_59.fit.gz')
    testpath = '2014/03/05/20140305_013000_59.fit.gz'
    d = parse_time((2014, 3, 5, 1, 30))
    assert s.matches(testpath, d)


def test_directory_date_patternFalse():
    s = Scraper(format='{{year:4d}}/{{month:2d}}/{{day:2d}}/{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}_59.fit.gz')
    testpath = '2013/03/05/20140305_013000_59.fit.gz'
    d = parse_time((2014, 3, 5, 1, 30))
    assert not s.matches(testpath, d)


def test_directory_obs_pattern():
    s = Scraper(format='{{year:2d}}{{month:2d}}{{day:2d}}/{observatory}_{{year:4d}}{{month:2d}}{{day:2d}}.fits', observatory='SDO')
    testpath = '140305/SDO_20140305.fits'
    d = parse_time((2014, 3, 5))
    assert s.matches(testpath, d)

def test_directory_range():
    s = Scraper(format='{{year:4d}}/{{month:2d}}/{{day:2d}}/{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}.fit.gz')
    directory_list = ['2009/12/30/', '2009/12/31/', '2010/01/01/',
                      '2010/01/02/', '2010/01/03/']
    timerange = TimeRange('2009-12-30', '2010-01-03')
    assert s.range(timerange) == directory_list


def test_directory_regex():
    # Test for Windows where '\' is a path separator and not part of the regex
    s = Scraper(format='scheme://a.url.with/a/few/forward/slashes/andbacklash\\inthename.ext')
    timerange = TimeRange('2019-02-01', '2019-02-03')
    directory = s.range(timerange)
    assert directory == ['scheme://a.url.with/a/few/forward/slashes/']


def test_directory_range_false():
    s = Scraper(format='{{year:4d}}{{month:2d}}{{day:2d}}/{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}.fit.gz')
    directory_list = ['20091230/', '20091231/', '20100101/',
                      '20090102/', '20090103/']
    timerange = TimeRange('2009/12/30', '2010/01/03')
    assert s.range(timerange) != directory_list


def testNoDateDirectory():
    s = Scraper(format='mySpacecraft/myInstrument/xMinutes/aaa{{year:4d}}{{month_name_abbr:l}}.ext')
    directory_list = ['mySpacecraft/myInstrument/xMinutes/']
    timerange = TimeRange('2009/11/20', '2010/01/03')
    assert s.range(timerange) == directory_list


def test_directory_range_hours():
    s = Scraper(format='{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}/{{hour:2d}}{{minute:2d}}.csv')
    timerange = TimeRange('2009-12-31T23:40:00', '2010-01-01T01:15:00')
    assert len(s.range(timerange)) == 3  # 3 directories (1 per hour)


def test_directory_range_single():
    s = Scraper(format='{{year:4d}}{{month:2d}}{{day:2d}}/{{hour:2d}}_{{minute:2d}}.csv')
    startdate = parse_time((2010, 10, 10, 5, 0))
    enddate = parse_time((2010, 10, 10, 7, 0))
    timerange = TimeRange(startdate, enddate)
    assert len(s.range(timerange)) == 1


def test_directory_range_month():
    s = Scraper(format='{{year:4d}}{{month:2d}}/{{day:2d}}/{{day_of_year:3d}}_{{hour:2d}}.txt')
    startdate = parse_time((2008, 2, 20, 10))
    enddate = parse_time((2008, 3, 2, 5))
    timerange = TimeRange(startdate, enddate)
    assert len(s.range(timerange)) == 12
    startdate = parse_time((2009, 2, 20, 10))
    enddate = parse_time((2009, 3, 2, 5))
    timerange = TimeRange(startdate, enddate)
    assert len(s.range(timerange)) == 11


@pytest.mark.parametrize(('pattern', 'filename', 'metadict'), [
    ('_{{year:4d}}{{month:2d}}{{day:2d}}__{{millisecond:3d}}c{{:5d}}_{{:2d}}{{}}.fts', '_20201535__012c12345_33 .fts',
     {'year': 2020, 'month': 15, 'day': 35, 'millisecond': 12}),
    ('fd_{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}_{{millisecond:3d}}.fts', 'fd_20130410_231211_119.fts',
     {'year': 2013, 'month': 4, 'day': 10, 'hour': 23, 'minute': 12, 'second': 11, 'millisecond': 119})
])
def test_url_pattern(pattern, filename, metadict):
    assert parse(pattern.format(None), filename).named == metadict


def test_url_pattern_milliseconds_zero_padded():
    # Asserts solution to ticket #1954.
    # Milliseconds must be zero-padded in order to match URL lengths.
    now_mock = Mock(return_value=datetime.datetime(2019, 4, 19, 0, 0, 0, 4009))
    with patch('sunpy.net.scraper.datetime', now=now_mock):
        s = Scraper(format='fd_{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}_{{millisecond:3d}}.fts')
    now_mock.assert_called_once()
    assert s.now == 'fd_20190419_000000_004.fts'


def test_files_range_same_directory_local():
    s = Scraper(format='/'.join(['file:/', str(rootdir),
                          'EIT_header', 'efz{{year:4d}}{{month:2d}}{{day:2d}}.{{hour:2d}}{{minute:2d}}{{second:2d}}_s.header']))
    startdate = parse_time((2004, 3, 1, 4, 0))
    enddate = parse_time((2004, 3, 1, 6, 30))
    assert len(s.filelist(TimeRange(startdate, enddate))) == 3
    startdate = parse_time((2010, 1, 10, 20, 30))
    enddate = parse_time((2010, 1, 20, 20, 30))
    assert len(s.filelist(TimeRange(startdate, enddate))) == 0


@pytest.mark.remote_data
def test_files_range_same_directory_remote():
    pattern = ('https://proba2.sidc.be/{instrument}/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/'
               '{instrument}_lv1_{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}.fits')
    s = Scraper(format=pattern, instrument='swap')
    startdate = parse_time((2014, 5, 14, 0, 0))
    enddate = parse_time((2014, 5, 14, 0, 5))
    timerange = TimeRange(startdate, enddate)
    assert len(s.filelist(timerange)) == 2
    startdate = parse_time((2014, 5, 14, 0, 6))
    enddate = parse_time((2014, 5, 14, 0, 7))
    timerange = TimeRange(startdate, enddate)
    assert len(s.filelist(timerange)) == 0


@pytest.mark.remote_data
def test_files_range_same_directory_months_remote():
    pattern = ('https://izw1.caltech.edu/{spacecraft}/DATA/{instrument}/'
               'Ahead/1minute/AeH{{year:2d}}{{month_name_abbr:w}}.1m')
    s = Scraper(format=pattern, spacecraft='STEREO', instrument='HET')
    startdate = parse_time((2007, 8, 1))
    enddate = parse_time((2007, 9, 10))
    timerange = TimeRange(startdate, enddate)
    files = s.filelist(timerange)
    assert files == ['https://izw1.caltech.edu/STEREO/DATA/HET/Ahead/1minute/AeH07Aug.1m',
                     'https://izw1.caltech.edu/STEREO/DATA/HET/Ahead/1minute/AeH07Sep.1m']


@pytest.mark.remote_data
def test_ftp():
    pattern = 'ftp://ftp.ngdc.noaa.gov/STP/space-weather/swpc-products/daily_reports/solar_region_summaries/{{year:4d}}/{{month:2d}}/{{year:4d}}{{month:2d}}{{day:2d}}SRS.txt'
    s = Scraper(format=pattern)
    timerange = TimeRange('2024/5/18', '2024/5/20')
    urls = s.filelist(timerange)
    assert urls[0] == ('ftp://ftp.ngdc.noaa.gov/STP/space-weather/swpc-products/daily_reports/solar_region_summaries/2024/05/20240518SRS.txt')
    assert len(urls) == 3


@pytest.mark.remote_data
def test_filelist_url_missing_directory():
    # Asserts solution to ticket #2684.
    # Attempting to access data for the year 1960 results in a 404, so no files are returned.
    pattern = 'https://lasp.colorado.edu/eve/data_access/evewebdata/products/level2/{{year:4d}}/{{day_of_year:3d}}/'
    s = Scraper(format=pattern)
    timerange = TimeRange('1960/01/01 00:00:00', '1960/01/02 00:00:00')
    assert len(s.filelist(timerange)) == 0


@pytest.mark.remote_data
def test_filelist_relative_hrefs():
    # the url opened by the scraper from below pattern contains some links which don't have hrefs
    pattern = 'https://www.bbso.njit.edu/pub/archive/{{year:4d}}/{{month:2d}}/{{day:2d}}/bbso_halph_fr_{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{minute:2d}}{{second:2d}}.fts'
    s = Scraper(format=pattern)
    timerange = TimeRange('2016/5/18 15:28:00', '2016/5/18 16:30:00')
    assert s.domain == 'https://www.bbso.njit.edu/'
    # hrefs are relative to domain here, not to the directory they are present in
    # this checks that `scraper.filelist` returns fileurls relative to the domain
    fileurls = s.filelist(timerange)
    assert fileurls[1] == s.domain + 'pub/archive/2016/05/18/bbso_halph_fr_20160518_160033.fts'


@pytest.mark.remote_data
def test_extract_files_meta():
    pattern0 = 'https://solar.nro.nao.ac.jp/norh/data/tcx/{{year:4d}}/{{month:2d}}/{{wave}}{{year:2d}}{{month:2d}}{{day:2d}}'
    s0 = Scraper(format=pattern0)
    timerange0 = TimeRange('2020/1/1 4:00', '2020/1/2')
    matchdict = {'wave': ['tca', 'tcz']}
    metalist0 = s0._extract_files_meta(timerange0, matcher=matchdict)
    assert metalist0[0]['wave'] == 'tca'
    assert metalist0[3]['wave'] == 'tcz'
    assert metalist0[1]['day'] == 2

    pattern1 = ('https://gong2.nso.edu/oQR/zqs/{{year:4d}}{{month:2d}}/mrzqs{{year:2d}}{{month:2d}}{{day:2d}}'
                '/mrzqs{{year:2d}}{{month:2d}}{{day:2d}}t{{hour:2d}}{{minute:2d}}c{{CAR_ROT:4d}}_{{:3d}}.fits.gz')
    s1 = Scraper(format=pattern1)
    timerange1 = TimeRange('2020-01-05', '2020-01-05T16:00:00')
    metalist1 = s1._extract_files_meta(timerange1)
    urls = s1.filelist(timerange1)
    assert metalist1[3]['CAR_ROT'] == 2226
    assert metalist1[-1]['url'] == urls[-1]


def test_no_directory():
    s = Scraper(format='files/{{year:4d}}{{month:2d}}{{day:2d}}_{{hour:2d}}{{month:2d}}.dat')
    startdate = parse_time((2010, 1, 10, 20, 30))
    enddate = parse_time((2010, 1, 20, 20, 30))
    timerange = TimeRange(startdate, enddate)
    assert len(s.range(timerange)) == 1


@pytest.mark.parametrize(('exdict', 'start', 'end'), [
    ({"year": 2000}, '2000-01-01 00:00:00', '2000-12-31 23:59:59.999000'),
    ({"year": 2016, "month": 2}, '2016-02-01 00:00:00', '2016-02-29 23:59:59.999000'),
    ({'year': 2019, 'month': 2, 'day': 28}, '2019-02-28 00:00:00', '2019-02-28 23:59:59.999000'),
    ({'year': 2020, 'month': 7, 'day': 31, 'hour': 23, 'minute': 59, 'second': 59},
     '2020-07-31 23:59:59', '2020-07-31 23:59:59.999000')])
def test_get_timerange_with_extractor(exdict, start, end):
    tr = TimeRange(start, end)
    file_timerange = get_timerange_from_exdict(exdict)
    assert file_timerange == tr


@pytest.mark.remote_data
def test_parse_pattern_data():
    prefix = 'https://gong2.nso.edu/oQR/zqs/'
    pattern = prefix + '{{year:4d}}{{month:2d}}/mrzqs{{year:2d}}{{month:2d}}{{day:2d}}/mrzqs{{year:2d}}{{month:2d}}{{day:2d}}t{{hour:2d}}{{minute:2d}}c{{:4d}}_{{:3d}}.fits.gz'
    s = Scraper(format=pattern)
    timerange = TimeRange('2020-01-05', '2020-01-06T16:00:00')
    assert len(s.filelist(timerange)) == 37


@pytest.mark.remote_data
def test_yearly_overlap():
    # Check that a time range that falls within the interval that a file represents
    # returns a single result.
    pattern = "https://www.ngdc.noaa.gov/stp/space-weather/solar-data/solar-features/solar-flares/x-rays/goes/xrs/goes-xrs-report_{{year:4d}}.txt"
    scraper = Scraper(format=pattern)

    # Should return a single file for 2013
    trange = TimeRange("2013-01-02", "2013-01-03")
    assert len(scraper.filelist(trange)) == 1


@pytest.mark.parametrize(
    ("error_code", "expected_number_calls", "error_message"),
    [
        (400, 1, "Bad Request"),
        (403, 1, "Forbidden"),
        (429, 6, "Too Many Requests"),
        (504, 6, "Gateway Timeout"),
    ],
)
def test_http_errors_with_enqueue_limit(error_code, expected_number_calls, error_message):
    with patch("sunpy.net.scraper.urlopen") as mocked_urlopen:
        mocked_urlopen.side_effect = HTTPError(
            "https://example.com", error_code, error_message, {}, None
        )
        time_range = TimeRange("2012/3/4", "2012/3/4 02:00")
        pattern = "https://proba2.sidc.be/lyra/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/{{}}_lev{{Level:1d}}_std.fits"
        scraper = Scraper(format=pattern)
        with pytest.raises(HTTPError, match=error_message) as excinfo:
            scraper._httpfilelist(time_range)
        assert excinfo.value.code == error_code
        assert mocked_urlopen.call_count == expected_number_calls


def test_connection_error():
    with patch('sunpy.net.scraper.urlopen') as mocked_urlopen:
        mocked_urlopen.side_effect = URLError('connection error')
        time = TimeRange('2012/3/4', '2012/3/4 02:00')
        pattern = "https://proba2.sidc.be/lyra/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/{{}}_lev{{Level:1d}}_std.fits"
        scraper = Scraper(format=pattern)
        with pytest.raises(URLError, match='connection error'):
            scraper._httpfilelist(time)


def test_http_404_error_debug_message(caplog):
    with caplog.at_level(logging.DEBUG, logger='sunpy'):
        def patch_range(self, range):
            return ['https://test.com/']
        with patch('sunpy.net.scraper.urlopen') as mocked_urlopen:
            with patch.object(Scraper, 'range', patch_range):
                mocked_urlopen.side_effect = HTTPError('https://example.com', 404, '', {}, None)
                time = TimeRange('2012/3/4', '2012/3/4 02:00')
                pattern = "https://proba2.sidc.be/lyra/data/bsd/{{year:4d}}/{{month:2d}}/{{day:2d}}/{{}}_lev{{Level:1d}}_std.fits"
                scraper = Scraper(format=pattern)
                scraper._httpfilelist(time)
                assert "Directory https://test.com/ not found." in caplog.text


def test_check_timerange():
    s = Scraper(format='{{year:4d}}.fits')
    # Valid time range for 2014.fits is the whole of 2014
    # Test different cases to make sure check_timerange is working as expected

    # Interval exactly on lower boundary
    assert s._check_timerange('2014.fits', TimeRange("2013-06-01", "2014-01-01"))
    # Overlaps lower boundary
    assert s._check_timerange('2014.fits', TimeRange("2013-06-01", "2014-01-02"))
    # Overlaps upper and lower boundary
    assert s._check_timerange('2014.fits', TimeRange("2013-06-01", "2015-01-02"))
    # Entirely within both boundaries
    assert s._check_timerange('2014.fits', TimeRange("2014-06-01", "2014-07-02"))
    # Overlaps upper boundary
    assert s._check_timerange('2014.fits', TimeRange("2014-06-01", "2015-01-02"))
    # Interval exactly on upper boundary
    assert not s._check_timerange('2014.fits', TimeRange("2015-01-01", "2015-01-02"))

    # Interval below both boundaries
    assert not s._check_timerange('2014.fits', TimeRange("2002-01-01", "2013-01-02"))
    # Interval above both boundaries
    assert not s._check_timerange('2014.fits', TimeRange("2022-01-01", "2025-01-02"))

    # Only 2-digit Year
    assert s._check_timerange('14.fits', TimeRange("2013-06-01", "2014-01-01"))


def test_local_expected_directory_doesnt_exist(tmp_path):
    path = (tmp_path / '2025' / '01' / '01')
    path.mkdir(parents=True)
    with (path / 'test.txt').open('w') as file:
        file.write('')
    s = Scraper(format='file://'+str(tmp_path)+'/{{year:4d}}/{{month:2d}}/{{day:2d}}/{{file}}')
    files = s.filelist(TimeRange("2025-01-01", "2025-01-02"))
    assert len(files) == 1



# Check if user has installed the visualisation extras
from sunpy.util.sysinfo import _warn_missing_deps

_warn_missing_deps('visualization')

from sunpy.visualization.visualization import *
from . import animator, colormaps, drawing



"""
====================================
Differentially rotating a coordinate
====================================

How to differentially rotate a coordinate.

The example uses the `~sunpy.coordinates.metaframes.RotatedSunFrame` coordinate
metaframe in `sunpy.coordinates` to apply differential rotation to a
coordinate. See :ref:`sunpy-topic-guide-coordinates-rotatedsunframe` for more details on
using `~sunpy.coordinates.metaframes.RotatedSunFrame`.
"""
import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates import RotatedSunFrame
from sunpy.data.sample import AIA_171_IMAGE

##############################################################################
# First, load an AIA observation and define a coordinate in its coordinate
# frame (here, helioprojective Cartesian). The appropriate rate of rotation
# is determined from the heliographic latitude of the coordinate.

aiamap = sunpy.map.Map(AIA_171_IMAGE)
point = SkyCoord(187*u.arcsec, 283*u.arcsec, frame=aiamap.coordinate_frame)

##############################################################################
# We can differentially rotate this coordinate by using
# `~sunpy.coordinates.metaframes.RotatedSunFrame` with an array of observation
# times. Let's define a daily cadence for +/- five days.

durations = np.concatenate([range(-5, 0), range(1, 6)]) * u.day
diffrot_point = SkyCoord(RotatedSunFrame(base=point, duration=durations))

##############################################################################
# To see what this coordinate looks like in "real" helioprojective
# Cartesian coordinates, we can transform it back to the original frame.
# Since these coordinates are represented in the original frame, they will not
# account for the changing position of the observer over this same time range.

transformed_diffrot_point = diffrot_point.transform_to(aiamap.coordinate_frame)
print(transformed_diffrot_point)

##############################################################################
# Let's plot the original coordinate and the differentially rotated
# coordinates on top of the AIA observation.

fig = plt.figure()
ax = fig.add_subplot(projection=aiamap)
aiamap.plot(axes=ax, clip_interval=(1., 99.95)*u.percent)

ax.plot_coord(point, 'ro', fillstyle='none', label='Original')
ax.plot_coord(transformed_diffrot_point, 'bo', fillstyle='none',
              label='Rotated')
ax.legend()

plt.show()



import sunpy.timeseries
from sunpy.data.test import get_test_filepath
from sunpy.tests.helpers import figure_test

lyra_filepath = get_test_filepath('lyra_20150101-000000_lev3_std_truncated.fits.gz')


def test_implicit_lyra():
    # Test a LYRA TimeSeries
    ts_lyra = sunpy.timeseries.TimeSeries(lyra_filepath)
    assert isinstance(ts_lyra, sunpy.timeseries.sources.lyra.LYRATimeSeries)


def test_lyra():
    # Test a LYRA TimeSeries
    ts_lyra = sunpy.timeseries.TimeSeries(lyra_filepath, source='LYRA')
    assert isinstance(ts_lyra, sunpy.timeseries.sources.lyra.LYRATimeSeries)


def test_lyra_plot_column(lyra_test_ts):
    axes = lyra_test_ts.plot(columns=['CHANNEL1', 'CHANNEL3'])
    assert len(axes) == 2
    assert 'Lyman alpha' in axes[0].get_ylabel()
    assert 'Al filter' in axes[1].get_ylabel()


@figure_test
def test_lyra_peek(lyra_test_ts):
    lyra_test_ts.peek()



"""
This module provides errors/exceptions and warnings of general use for SunPy.

Exceptions that are specific to a given package should **not** be here,
but rather in the particular package.
"""
import warnings

__all__ = ["NoMapsInFileError",
           "SunpyWarning", "SunpyUserWarning", "SunpyDeprecationWarning",
           "SunpyPendingDeprecationWarning", "SunpyMetadataWarning", "SunpyConnectionWarning",
           "warn_user", "warn_deprecated", "warn_metadata", "warn_connection"]


class NoMapsInFileError(Exception):
    """
    An error raised when a file is opened and no maps are found.
    """


class SunpyWarning(Warning):
    """
    The base warning class from which all Sunpy warnings should inherit.

    Any warning inheriting from this class is handled by the Sunpy
    logger. This warning should not be issued in normal code. Use
    "SunpyUserWarning" instead or a specific sub-class.
    """


class SunpyUserWarning(UserWarning, SunpyWarning):
    """
    The primary warning class for Sunpy.

    Use this if you do not need a specific type of warning.
    """


class SunpyMetadataWarning(UserWarning, SunpyWarning):
    """
    Warning class for cases metadata is missing.

    This does not inherit from SunpyWarning because we want to use
    stacklevel=3 to show the user where the issue occurred in their code.
    """


class SunpyDeprecationWarning(FutureWarning, SunpyWarning):
    """
    A warning class to indicate a deprecated feature.
    """


class SunpyPendingDeprecationWarning(PendingDeprecationWarning, SunpyWarning):
    """
    A warning class to indicate a soon-to-be deprecated feature.
    """


class SunpyConnectionWarning(SunpyUserWarning):
    """
    A warning class to indicate a connection warning.

    This warning should be issued when a recoverable error occurs during a connection to a remote server, such as falling back to a mirror etc.

    This will not fail the CI (via a pytest ignore) as it is not a critical warning.
    """


def warn_metadata(msg, stacklevel=1):
    """
    Raise a `SunpyMetadataWarning`.

    Parameters
    ----------
    msg : str
        Warning message.
    stacklevel : int
        This is interpreted relative to the call to this function,
        e.g. ``stacklevel=1`` (the default) sets the stack level in the
        code that calls this function.
    """
    warnings.warn(msg, SunpyMetadataWarning, stacklevel + 1)


def warn_user(msg, stacklevel=1):
    """
    Raise a `SunpyUserWarning`.

    Parameters
    ----------
    msg : str
        Warning message.
    stacklevel : int
        This is interpreted relative to the call to this function,
        e.g. ``stacklevel=1`` (the default) sets the stack level in the
        code that calls this function.
    """
    warnings.warn(msg, SunpyUserWarning, stacklevel + 1)


def warn_connection(msg, stacklevel=1):
    """
    Raise a `SunpyConnectionWarning`.

    Parameters
    ----------
    msg : str
        Warning message.
    stacklevel : int
        This is interpreted relative to the call to this function,
        e.g. ``stacklevel=1`` (the default) sets the stack level in the
        code that calls this function.
    """
    warnings.warn(msg, SunpyConnectionWarning, stacklevel + 1)


def warn_deprecated(msg, stacklevel=1):
    """
    Raise a `SunpyDeprecationWarning`.

    Parameters
    ----------
    msg : str
        Warning message.
    stacklevel : int
        This is interpreted relative to the call to this function,
        e.g. ``stacklevel=1`` (the default) sets the stack level in the
        code that calls this function.
    """
    warnings.warn(msg, SunpyDeprecationWarning, stacklevel + 1)



"""PROBA2 Map subclass definitions"""
from sunpy.map import GenericMap

__all__ = ['SWAPMap']


class SWAPMap(GenericMap):
    """PROBA2 SWAP Image Map.

    The Sun Watcher using Active Pixel System detector and Image Processing (SWAP)
    SWAP provides images of the solar corona at about 17.4 nm, a bandpass
    that corresponds to a temperature of roughly 1 million degrees,
    with a cadence of 1 image per 1-2 minutes, and field of view (FOV) of 54 arcmin.
    It is derived from the SOHO EIT telescope concept design.

    PROBA2 was launched on 2 November 2009.

    References
    ----------
    * `Proba2 SWAP Science Center <https://proba2.sidc.be/index.php>`__
    * `Fits headers reference <https://proba2.sidc.be/data/SWAP/level0>`__

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset

        import sunpy.map
        import astropy.units as u
        from sunpy.data.sample import SWAP_LEVEL1_IMAGE
        swap_map = sunpy.map.Map(SWAP_LEVEL1_IMAGE)
        swap_map.plot(clip_interval=(1, 99)*u.percent)
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self._nickname = self.detector
        self.plot_settings['cmap'] = 'sdoaia171'

    @property
    def observatory(self):
        return "PROBA2"

    @property
    def detector(self):
        return "SWAP"

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an SWAP image"""
        return header.get('instrume') == 'SWAP'



"""STEREO Map subclass definitions"""
import astropy.units as u
from astropy.visualization import PowerStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.map import GenericMap
from sunpy.map.sources.source_type import source_stretch

__all__ = ['EUVIMap', 'CORMap', 'HIMap']


class EUVIMap(GenericMap):
    """STEREO-SECCHI EUVI Image Map

    EUVI is an extreme ultraviolet (EUV) imager. Part of the STEREO-SECCHI
    suite it observes the Sun from 1 to 1.7 solar radii. It is capable of
    observing at 304 (He II), 171 (Fe IX), 195 (Fe XII), and 284 (Fe XV)
    Angstroms.

    References
    ----------
    * `STEREO Mission Page <https://stereo.gsfc.nasa.gov/>`__
    * `STEREO SECCHI <https://secchi.nrl.navy.mil/>`__
    * `Instrument Page <https://secchi.lmsal.com/EUVI/>`__

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset

        import sunpy.map
        from sunpy.data.sample import EUVI_STEREOSCOPIC_IMAGE
        euvi_map = sunpy.map.Map(EUVI_STEREOSCOPIC_IMAGE)
        euvi_map.plot()
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self._nickname = f"{self.detector}-{self.observatory[-1]}"
        self.plot_settings['cmap'] = f'euvi{int(self.wavelength.value):d}'
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.25)), clip=False)

    def _rotation_matrix_from_crota(self):
        return super()._rotation_matrix_from_crota('CROTA')

    @property
    def waveunit(self):
        unit = self.meta.get("waveunit", "Angstrom")
        return u.Unit(unit)

    @property
    def rsun_arcseconds(self):
        return self.meta.get('rsun', None)

    @property
    def rsun_obs(self):
        rsun_arcseconds = self.rsun_arcseconds

        if rsun_arcseconds is None:
            rsun_arcseconds = super().rsun_obs

        return u.Quantity(rsun_arcseconds, 'arcsec')

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an EUVI image"""
        return header.get('detector') == 'EUVI'


class CORMap(GenericMap):
    """STEREO-SECCHI CORonograph Image Map.

    Part of the STEREO-SECCHI suite of remote sensing telescopes,
    COR is a set of two coronographs (COR1, COR2) onboard STEREO.
    They are both traditional Lyot coronagraphs.

    The COR1 detectors observes from 1.3 to 4 solar radii while the
    COR2 detectors observe a range from 2 to 15 solar radii.

    References
    ----------
    * `STEREO Mission Page <https://stereo.gsfc.nasa.gov/>`__
    * `STEREO SECCHI <https://secchi.nrl.navy.mil/>`__
    * `COR1 Instrument Page <https://cor1.gsfc.nasa.gov>`__
    * `COR2 Instrument Page <https://secchi.nrl.navy.mil//index.php?p=cor2>`__
    * `COR1 User Guide <https://cor1.gsfc.nasa.gov/guide/>`__
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self._nickname = f"{self.detector}-{self.observatory[-1]}"
        self.plot_settings['cmap'] = f'stereocor{self.detector[-1]!s}'
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.5)), clip=False)

    @property
    def measurement(self):
        # TODO: This needs to do more than white-light. Should give B, pB, etc.
        return "white-light"

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an COR image"""
        return str(header.get('detector', '')).startswith('COR')


class HIMap(GenericMap):
    """STEREO-SECCHI Heliospheric Imager (HI) Map.

    The HI is a wide-angle visible-light imaging system
    for the detection of coronal mass ejection (CME) events
    in interplanetary space and, in particular, of events
    directed towards the Earth.

    The Heliospheric imager consists of two instruments, the HI-1 and HI-2.
    The HI1 observes from 15-80 solar radii while HI2 observes from 80-215
    solar radii.

    References
    ----------
    * `STEREO Mission Page <https://stereo.gsfc.nasa.gov/>`__
    * `STEREO SECCHI <https://secchi.nrl.navy.mil>`__
    * `HI Instrument Page <https://www.stereo.rl.ac.uk/>`__
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self._nickname = f"{self.detector}-{self.observatory[-1]}"
        self.plot_settings['cmap'] = f'stereohi{self.detector[-1]!s}'
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.25)), clip=False)

    @property
    def measurement(self):
        # TODO: This needs to do more than white-light. Should give B, pB, etc.
        return "white-light"

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an COR image"""
        return str(header.get('detector', '')).startswith('HI')



"""
========================
Creating Carrington Maps
========================

In this example we use the `reproject` generate a map in heliographic Carrington coordinates from a full-disk AIA image.

You will need `reproject <https://reproject.readthedocs.io/en/stable/>`__ v0.6 or higher installed.
"""
# sphinx_gallery_thumbnail_number = 2

import matplotlib.pyplot as plt

import sunpy.data.sample
import sunpy.map
from sunpy.map.header_helper import make_heliographic_header

###############################################################################
# We will start with using sunpy's sample data for this example.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_193_IMAGE)

fig = plt.figure()
ax = fig.add_subplot(projection=aia_map)
aia_map.plot(axes=ax)

###############################################################################
# Reproject works by transforming an input image to a desired World Coordinate
# System (WCS) projection. Here we use :func:`sunpy.map.header_helper.make_heliographic_header`
# to create a FITS WCS header based on a heliographic Carrington reference
# coordinate.

shape = (720, 1440)
carr_header = make_heliographic_header(aia_map.date, aia_map.observer_coordinate, shape, frame='carrington')

###############################################################################
# With the new header, re-project the data into the new coordinate system.
# The :meth:`~sunpy.map.GenericMap.reproject_to` defaults to using
# the fast :func:`reproject.reproject_interp` algorithm, but a different
# algorithm can be specified (e.g., :func:`reproject.reproject_adaptive`).

outmap = aia_map.reproject_to(carr_header)

###############################################################################
# Plot the result.

fig = plt.figure()
ax = fig.add_subplot(projection=outmap)
outmap.plot(axes=ax)
outmap.draw_limb(color='blue')

plt.show()



"""
===============
Resampling Maps
===============

How to resample a map using the resample method, which implements interpolation, or
using superpixels, which combines pixels.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
import sunpy.map

##############################################################################
# We start with the sample data.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)

##############################################################################
# To reduce the angular resolution of the map, you can use the
# :meth:`~sunpy.map.GenericMap.resample` method, specifying the new dimensions
# in pixels. By default, this method uses linear interpolation but this can be
# changed with the ``method`` argument ('nearest', 'linear' or 'spline').

new_dimensions = [40, 40] * u.pixel
aia_resampled_map = aia_map.resample(new_dimensions)

##############################################################################
# Let's plot the result.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_resampled_map)
aia_resampled_map.plot(axes=ax)
plt.show()

##############################################################################
# Another way to reduce the angular resolution of the map is by using the
# :meth:`~sunpy.map.GenericMap.superpixel` method, which combines pixels.
# The superpixel dimensions do not need to be square, and the intensity of
# each superpixel defaults to the sum of the constituent pixels. For example,
# you can reduce the AIA map resolution by a factor of 16 by specifying 16x16
# superpixels.

superpixel_size = [16, 16] * u.pixel
aia_superpixel_map = aia_map.superpixel(superpixel_size)

##############################################################################
# Let's plot the result.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_superpixel_map)
aia_superpixel_map.plot(axes=ax)
plt.show()



import re

import pytest

from sunpy.net import base_client, dataretriever, jsoc, vso
from sunpy.net.base_client import QueryResponseTable, convert_row_to_table
from sunpy.net.dataretriever.sources.norh import NoRHClient

_REGEX = re.compile(r"Client")

CLIENT_LIST = []

for a_import in [vso, jsoc, dataretriever]:
    for item in dir(a_import):
        if _REGEX.search(item):
            CLIENT_LIST.append(getattr(a_import, item))

CLIENT_LIST.remove(dataretriever.client.GenericClient)

# We can access the registry directly
CLIENT_NAMES = base_client.BaseClient._registry.keys()
CLIENTS_REG = base_client.BaseClient._registry.items()


@pytest.mark.parametrize("client", CLIENT_LIST)
def test_registry(client):
    """
    Check if each client has been registered.
    """
    assert client in CLIENT_NAMES
    assert (client, client._can_handle_query) in CLIENTS_REG


@pytest.fixture
def dummy_response():
    return QueryResponseTable([{'hello': 1}], client=NoRHClient())


def test_slice(dummy_response):
    assert len(dummy_response) == 1

    row = dummy_response[0]
    table = row.as_table()

    assert len(table) == 1
    assert isinstance(table.client, NoRHClient)

    col = dummy_response['hello']
    table = col.as_table()

    assert len(table) == 1
    assert isinstance(table.client, NoRHClient)


def test_path_format_keys(dummy_response):
    assert dummy_response.path_format_keys() == {'hello'}


def test_convert_row_to_table(dummy_response):

    @convert_row_to_table
    def example(self, query_results, **kwargs):
        return query_results

    assert example(None, dummy_response) is dummy_response
    # This is a single row table anyway
    assert example(None, dummy_response[0]) == dummy_response



import re

import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.wcs import WCS

import sunpy.map
from sunpy.coordinates import frames, sun
from sunpy.map import make_fitswcs_header
from sunpy.map.header_helper import make_heliographic_header, make_hpr_header
from sunpy.util.metadata import MetaDict


@pytest.fixture
def map_data():
    return np.random.rand(20, 10)


@pytest.fixture
def hpc_coord():
    return SkyCoord(0*u.arcsec, 100*u.arcsec, observer='earth', rsun=7.1e8*u.m,
                    obstime='2013-10-28 00:00', frame=frames.Helioprojective)


@pytest.fixture
def hgc_coord():
    return SkyCoord(70*u.deg, -30*u.deg, 1.5*u.AU, observer='self', rsun=7.2e8*u.m,
                    obstime='2013-10-28 00:00', frame=frames.HeliographicCarrington)


@pytest.fixture
def hgs_coord():
    return SkyCoord(-50*u.deg, 50*u.deg, rsun=7.3e8*u.m,
                    obstime='2013-10-28 00:00', frame=frames.HeliographicStonyhurst)


@pytest.fixture
def hcc_coord():
    return SkyCoord(-72241*u.km, 361206.1*u.km, 589951.4*u.km, observer='earth',
                    obstime='2013-10-28 00:00', frame=frames.Heliocentric)


@pytest.fixture
def hpc_header(map_data, hpc_coord):
    return make_fitswcs_header(map_data, hpc_coord)


@pytest.fixture
def hgc_header(map_data, hgc_coord):
    return make_fitswcs_header(map_data, hgc_coord, projection_code='CAR')


@pytest.fixture
def hgs_header(map_data, hgs_coord):
    return make_fitswcs_header(map_data, hgs_coord, projection_code='CAR')


@pytest.fixture
def hpc_coord_notime():
    return SkyCoord(0*u.arcsec, 0*u.arcsec, frame=frames.Helioprojective)


def test_scale_conversion(map_data, hpc_coord):
    # The header will have cunit1/2 of arcsec
    header = make_fitswcs_header(map_data, hpc_coord, scale=[1, 2] * u.arcmin / u.pix)
    assert header['cdelt1'] == 60
    assert header['cdelt2'] == 120


def test_default_rotation(map_data, hpc_coord):
    header = make_fitswcs_header(map_data, hpc_coord)
    wcs = WCS(header)
    np.testing.assert_allclose(wcs.wcs.pc, [[1, 0], [0, 1]], atol=1e-5)


def test_rotation_angle(map_data, hpc_coord):
    header = make_fitswcs_header(map_data, hpc_coord,
                                 rotation_angle=90*u.deg)
    wcs = WCS(header)
    np.testing.assert_allclose(wcs.wcs.pc, [[0, -1], [1, 0]], atol=1e-5)


def test_rotation_angle_rectangular_pixels(map_data, hpc_coord):
    header = make_fitswcs_header(map_data, hpc_coord, scale=[2, 5] * u.arcsec / u.pix,
                                 rotation_angle=45*u.deg)
    wcs = WCS(header)
    np.testing.assert_allclose(wcs.wcs.pc, np.sqrt(0.5) * np.array([[1, -2.5], [0.4, 1]]), atol=1e-5)


def test_rotation_matrix(map_data, hpc_coord):
    header = make_fitswcs_header(map_data, hpc_coord,
                                 rotation_matrix=np.array([[1, 0], [0, 1]]))
    wcs = WCS(header)
    np.testing.assert_allclose(wcs.wcs.pc, [[1, 0], [0, 1]], atol=1e-5)


def test_hpc_header(hpc_header, hpc_coord):
    assert isinstance(hpc_header, MetaDict)

    assert hpc_header['naxis1'] == 10
    assert hpc_header['naxis2'] == 20
    assert hpc_header['crval1'] == 0
    assert hpc_header['crpix1'] == 5.5
    assert hpc_header['ctype1'] == 'HPLN-TAN'
    assert hpc_header['crval2'] == 100.
    assert hpc_header['crpix2'] == 10.5
    assert hpc_header['ctype2'] == 'HPLT-TAN'

    assert hpc_header['lonpole'] == 180.
    assert u.allclose(hpc_header['rsun_ref'] * u.m, hpc_coord.rsun)

    # Check for observer info for HPC
    assert u.allclose(hpc_header['hgln_obs'] * u.deg, hpc_coord.observer.lon)
    assert u.allclose(hpc_header['hglt_obs'] * u.deg, hpc_coord.observer.lat)
    assert u.allclose(hpc_header['dsun_obs'] * u.m, hpc_coord.observer.radius)
    assert u.allclose(hpc_header['rsun_obs'] * u.arcsec,
                      sun._angular_radius(hpc_coord.rsun, hpc_coord.observer.radius))

    assert isinstance(WCS(hpc_header), WCS)


def test_hgc_header(hgc_header, hgc_coord):
    assert isinstance(hgc_header, MetaDict)

    assert hgc_header['naxis1'] == 10
    assert hgc_header['naxis2'] == 20
    assert hgc_header['crval1'] == 70
    assert hgc_header['crpix1'] == 5.5
    assert hgc_header['ctype1'] == "CRLN-CAR"
    assert hgc_header['crval2'] == -30
    assert hgc_header['crpix2'] == 10.5
    assert hgc_header['ctype2'] == "CRLT-CAR"
    assert hgc_header['cunit1'] == "deg"
    assert hgc_header['cunit2'] == "deg"

    assert hgc_header['lonpole'] == 180.  # for negative reference latitude in a CAR projection
    assert u.allclose(hgc_header['rsun_ref'] * u.m, hgc_coord.rsun)

    # Check for observer info for HGC (which is set to "self")
    assert u.allclose(hgc_header['crln_obs'] * u.deg, hgc_coord.lon)
    assert u.allclose(hgc_header['crlt_obs'] * u.deg, hgc_coord.lat)
    assert u.allclose(hgc_header['dsun_obs'] * u.m, hgc_coord.radius)
    assert u.allclose(hgc_header['rsun_obs'] * u.arcsec,
                      sun._angular_radius(hgc_coord.rsun, hgc_coord.radius))

    assert isinstance(WCS(hgc_header), WCS)


def test_hgs_header(hgs_header, hgs_coord):
    assert isinstance(hgs_header, MetaDict)

    assert hgs_header['naxis1'] == 10
    assert hgs_header['naxis2'] == 20
    assert hgs_header['crval1'] == -50
    assert hgs_header['crpix1'] == 5.5
    assert hgs_header['ctype1'] == "HGLN-CAR"
    assert hgs_header['crval2'] == 50
    assert hgs_header['crpix2'] == 10.5
    assert hgs_header['ctype2'] == "HGLT-CAR"
    assert hgs_header['cunit1'] == "deg"
    assert hgs_header['cunit2'] == "deg"

    assert hgs_header['lonpole'] == 0.  # for positive reference latitude in a CAR projection
    assert u.allclose(hgs_header['rsun_ref'] * u.m, hgs_coord.rsun)

    # Check no observer info for HGS
    assert "hgln_obs" not in hgs_header
    assert "hglt_obs" not in hgs_header
    assert 'dsun_obs' not in hgs_header
    assert 'rsun_obs' not in hgs_header

    assert isinstance(WCS(hgs_header), WCS)


def test_instrument_keyword(map_data, hpc_coord):
    instrument_kwargs = {
        'instrument': 'test name',
        'observatory': 'test observatory',
        'telescope': 'test telescope',
        'detector': 'test detector',
        'wavelength': 171 * u.Angstrom,
        'exposure': 2 * u.s,
        'unit': u.Unit('ct s-1'),
    }
    header = make_fitswcs_header(map_data, hpc_coord, **instrument_kwargs)
    assert header['instrume'] == instrument_kwargs['instrument']
    assert header['obsrvtry'] == instrument_kwargs['observatory']
    assert header['telescop'] == instrument_kwargs['telescope']
    assert header['detector'] == instrument_kwargs['detector']
    assert header['wavelnth'] == instrument_kwargs['wavelength'].to_value()
    assert header['waveunit'] == instrument_kwargs['wavelength'].unit.to_string("fits")
    assert header['exptime'] == instrument_kwargs['exposure'].to_value('s')
    assert header['bunit'] == instrument_kwargs['unit'].to_string("fits")

    # Check returned MetaDict will make a `sunpy.map.Map`
    map_test = sunpy.map.Map(map_data, header)
    assert isinstance(map_test, sunpy.map.mapbase.GenericMap)


def test_quantity_input(map_data, hpc_coord):
    # Test that unit information is extracted correctly when data array is a quantity
    map_unit = u.Unit('ct / s')
    map_quantity = u.Quantity(map_data, map_unit)
    header = make_fitswcs_header(map_quantity, hpc_coord)
    assert header['bunit'] == map_unit.to_string('fits')
    # In cases where unit is specified and data is a quantity, specified unit will take
    # precedence
    override_unit = u.Unit('erg cm-2 s-1')
    header = make_fitswcs_header(map_quantity, hpc_coord, unit=override_unit)
    assert header['bunit'] == override_unit.to_string('fits')


def test_unit_as_string(map_data, hpc_coord):
    # Test that unit can be passed in as a string
    map_unit = u.Unit('ct / (pix s)')
    header = make_fitswcs_header(map_data, hpc_coord, unit=map_unit.to_string())
    assert header['bunit'] == map_unit.to_string(format='fits')


@pytest.mark.parametrize(('input_unit','output_string'),
                         [('DN', 'DN'), (u.DN, 'DN'), ('DN / s', 'DN / s')])
def test_make_fitswcs_header_handles_dn(input_unit, output_string, map_data, hpc_coord):
    header = make_fitswcs_header(map_data, hpc_coord, unit=input_unit)
    assert header['bunit'] == output_string

@pytest.mark.parametrize(
    ("coordinate_input", "expected_error_message"),
    [
        ("hcc_coord", "This function does not currently support heliocentric coordinates."),
        ("map_data", "coordinate needs to be a coordinate frame or an SkyCoord instance."),
        ("hpc_coord_notime", "The coordinate needs an observation time, `obstime`.")
    ]
)
def test_invalid_inputs(coordinate_input, expected_error_message, map_data, hcc_coord, hpc_coord_notime, hpc_coord):
    coordinates = {
        "hcc_coord": hcc_coord,
        "map_data": map_data,
        "hpc_coord_notime": hpc_coord_notime,
    }
    coord = coordinates[coordinate_input]


    with pytest.raises(ValueError, match=re.escape(expected_error_message)):
        make_fitswcs_header(map_data, coord)

    # Check arguments not given as astropy Quantities
    with pytest.raises(TypeError):
        make_fitswcs_header(map_data, hpc_coord, reference_pixel=[0, 0])
    with pytest.raises(TypeError):
        make_fitswcs_header(map_data, hpc_coord, scale=[0, 0])

    # Check arguments of reference_pixel and scale have to be given in astropy units of pix, and arcsec/pix
    with pytest.raises(u.UnitsError):
        make_fitswcs_header(map_data, hpc_coord, reference_pixel=u.Quantity([0, 0]))
    with pytest.raises(u.UnitsError):
        make_fitswcs_header(map_data, hpc_coord, scale=u.Quantity([0, 0]))
    with pytest.raises(u.UnitsError):
        make_fitswcs_header(map_data, hpc_coord, scale=u.Quantity([0, 0]*u.arcsec))


@pytest.mark.parametrize('frame', ['carrington', 'stonyhurst'])
# Second case here chosen to produce non-square pixels
@pytest.mark.parametrize('shape', [[90, 180], [240, 100]])
@pytest.mark.parametrize('projection_code', ['CAR', 'CEA'])
def test_make_heliographic_header(aia171_test_map, shape, projection_code, frame):
    header = make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, shape, frame=frame, projection_code=projection_code)
    carr_map = aia171_test_map.reproject_to(header)

    # Check upper right and lower left coordinates are as expected
    ll_coord = carr_map.pixel_to_world(-0.5 * u.pix, -0.5 * u.pix)
    assert ll_coord.lon in [-180 * u.deg, 180*u.deg]
    assert ll_coord.lat == -90 * u.deg
    assert ll_coord.frame.name == f"heliographic_{frame}"

    ur_coord = carr_map.pixel_to_world((shape[1] - 0.5) * u.pix, (shape[0] - 0.5) * u.pix)
    assert ur_coord.lon in [-180 * u.deg, 180*u.deg]
    assert ur_coord.lat == 90 * u.deg
    assert ur_coord.frame.name == f"heliographic_{frame}"


def test_make_heliographic_header_invalid_inputs(aia171_test_map):
    with pytest.raises(ValueError, match='projection_code must be one of'):
        make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington', projection_code='blah')

    with pytest.raises(ValueError, match='frame must be one of'):
        make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='blah')

    with pytest.raises(TypeError):
        make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington', map_center_longitude=0)

    with pytest.raises(u.UnitsError):
        make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington', map_center_longitude=0*u.m)

    # Test new keyword propagates correctly to header reference pixel coordinate
    # hgc/hgs = heliographic carrington/stonyhurst
    header_test_hgs = make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington')
    header_test_hgc = make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington', map_center_longitude=180*u.deg)
    assert header_test_hgs['crval1'] == 0.0
    assert header_test_hgc['crval1'] == 180.0
    # Test keyword wraps correctly to range [0,360]
    header_test_above = make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington', map_center_longitude=361*u.deg)
    header_test_below = make_heliographic_header(aia171_test_map.date, aia171_test_map.observer_coordinate, [90, 180], frame='carrington', map_center_longitude=-1*u.deg)
    assert header_test_above['crval1'] == 1.0
    assert header_test_below['crval1'] == 359.0


def test_make_hpr_header(hgs_coord):
    header = make_hpr_header(hgs_coord, (100, 720), theta_binsize=0.1*u.deg)
    assert header['ctype1'] == 'HRLN-CAR'
    assert header['ctype2'] == 'HRLT-CAR'
    assert header['naxis1'] == 720
    assert header['naxis2'] == 100
    assert header['crpix1'] == 360.5
    assert header['crpix2'] == 900.5
    assert header['crval1'] == 180.0
    assert header['crval2'] == 0.0
    assert header['hgln_obs'] == -50.0
    assert header['hglt_obs'] == 50.0
    assert Time(header['date-obs']) == hgs_coord.obstime



"""
Parker Solar Probe subclass definitions.
"""

from astropy.visualization import AsinhStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.map import GenericMap

__all__ = ['WISPRMap']


class WISPRMap(GenericMap):
    """
    WISPR Map

    The The Wide-field Imager for Parker Solar Probe (WISPR) is a white light
    telescope onboard the Parker Solar Probe (PSP) spacecraft.

    Notes
    -----
    By default, plotting of this map will set the lower bound to zero
    (i.e., clip out negative values for pixels). You can change this bound
    by modifying ``.plot_settings['norm'].vmin``.

    References
    ----------
    * `PSP science gateway <https://sppgway.jhuapl.edu//>`__
    * `WISPR Instrument Page <https://wispr.nrl.navy.mil//>`__
    * :cite:t:`vourlidas_wide-field_2016`
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self.plot_settings['norm'] = ImageNormalize(
            stretch=AsinhStretch(a=0.001), vmin=0)

    @property
    def processing_level(self):
        lvl = self.meta.get('level', None)
        if lvl is None:
            return
        # Chop off the leading 'L' if present
        if lvl[0] == 'L':
            lvl = lvl[1:]
        try:
            lvl = int(lvl)
        except ValueError:
            # The int conversion will fail for L2b files, and we should fail
            # safe if the user chooses to customize this with other values.
            pass
        return lvl

    @property
    def name(self):
        return 'WISPR ' + super().name

    @property
    def detector(self):
        detector = self.meta.get('detector', "")
        if detector == 1:
            return "Inner"
        if detector == 2:
            return "Outer"
        # Official data products will only be 1 or 2, but we should fail safe
        # if users customize this value themselves.
        return detector

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an WISPR image"""
        is_psp = 'parker solar probe' in str(header.get('obsrvtry', '')).lower()
        is_wispr = str(header.get('instrume', '')).startswith('WISPR')
        return is_psp and is_wispr



import pytest

import astropy.units as u
from astropy.table import Column, MaskedColumn, Table
from astropy.tests.helper import assert_quantity_allclose

from sunpy.data.test import get_test_filepath
from sunpy.io.special import srs

filenames = [{'file': 'SRS/20150906SRS.txt', 'rows': 5},
             {'file': 'SRS/20150306SRS.txt', 'rows': 4},
             {'file': 'SRS/20150101SRS.txt', 'rows': 9},
             {'file': 'SRS/20100621SRS.txt', 'rows': 3},  # This is a corrected copy
             # Uppercase files
             {'file': 'SRS/19960106SRS.txt', 'rows': 4},  # inc. spurious `NNN` on final line
             {'file': 'SRS/19960430SRS.txt', 'rows': 1},
             {'file': 'SRS/19960513SRS.txt', 'rows': 4},  # inc. empty `COMMENT` column
             {'file': 'SRS/20000922SRS.txt', 'rows': 10},  # inc. line with `III.`
             {'file': 'SRS/20000927SRS.txt', 'rows': 9},  # inc. line with `EFFECTIVE 2 OCT`
             {'file': 'SRS/20001001SRS.txt', 'rows': 12},  # inc. line with `COMMENT`
             {'file': 'SRS/20020624SRS.txt', 'rows': 14},  # inc. line with `PLAIN`
             {'file': 'SRS/20020628SRS.txt', 'rows': 13}]  # inc. line with `This message`

COORDINATES = [{'text': 'N10W05', 'latitude': 10, 'longitude': 5},
               {'text': 'N89E00', 'latitude': 89, 'longitude': 0},
               {'text': 'S33E02', 'latitude': -33, 'longitude': -2},
               {'text': 'S01', 'latitude': -1, 'longitude': None}]

LOCATION = Column(data=[x['text'] for x in COORDINATES], name='Location')
LONGLAT = Table()
LONGLAT.add_column(MaskedColumn(data=[x['longitude'] for x in COORDINATES], name='Longitude',
                                unit=u.deg, mask=True, fill_value=-99999))
LONGLAT.add_column(MaskedColumn(data=[x['latitude'] for x in COORDINATES], name='Latitude',
                                unit=u.deg, fill_value=-99999))


@pytest.mark.filterwarnings('ignore:dropping mask in Quantity column')
@pytest.mark.parametrize(
    ('path', 'number_of_rows'),
    [(get_test_filepath(elem['file']), elem['rows']) for elem in filenames]
)
def test_number_of_rows(path, number_of_rows):
    table = srs.read_srs(path)
    assert len(table) == number_of_rows


@pytest.mark.parametrize(('text', 'longitude'),
                         [(elem['text'], elem['longitude']) for elem in COORDINATES])
def test_parse_longitude(text, longitude):
    assert srs.parse_longitude(text) == longitude


@pytest.mark.parametrize(('text', 'latitude'),
                         [(elem['text'], elem['latitude']) for elem in COORDINATES])
def test_parse_latitude(text, latitude):
    assert srs.parse_latitude(text) == latitude


@pytest.mark.parametrize(('loc_column', 'exp_longitude', 'exp_latitude'),
                         [(LOCATION, LONGLAT['Longitude'], LONGLAT['Latitude'])])
def test_parse_location(loc_column, exp_longitude, exp_latitude):
    latitude, longitude = srs.parse_location(loc_column)
    assert_quantity_allclose(latitude, exp_latitude)
    assert_quantity_allclose(longitude, exp_longitude)

# test _try_drop_empty_column (tdec)


@pytest.fixture
def data_lines():
    return ['NMBR  LOCATION  LO  COMMENT', '8000  S14W96   232']


@pytest.fixture
def expected_pattern_dict():
    return {
        'Nmbr': r'^\d+$',
        'Location': r'^(?:[NESW](?:\d{2})){1,2}$',
        'Lo': r'^\d+$',
        'Comment': r'^[a-zA-Z]+$',
        'Additional Number': r'^\d+$',
    }


@pytest.fixture
def column_name_to_drop():
    return 'COMMENT'


def test_tdec(data_lines, expected_pattern_dict, column_name_to_drop):
    result = srs._try_drop_empty_column(column_name_to_drop, data_lines, expected_pattern_dict)
    out = data_lines.copy()
    out[0] = ' '.join([col.title() for col in out[0].split()][:-1])
    assert result == out


def test_tdec_smallest_dict_example(data_lines, expected_pattern_dict, column_name_to_drop):
    """
    Smallest possible dictionary given the ``data_lines``
    """
    keys_to_keep = ['Nmbr', 'Location', 'Lo']
    filtered_dict = {key: value for key, value in expected_pattern_dict.items() if key in keys_to_keep}
    result = srs._try_drop_empty_column(column_name_to_drop, data_lines, filtered_dict)
    out = data_lines.copy()
    out[0] = ' '.join([col.title() for col in out[0].split()][:-1])
    assert result == out


def test_tdec_too_small_dict_example(data_lines, expected_pattern_dict, column_name_to_drop):
    """
    Columns aren't a subset of the dictionary: no pattern to match for column `LO`
    """
    keys_to_keep = ['Nmbr', 'Location']
    filtered_dict = {key: value for key, value in expected_pattern_dict.items() if key in keys_to_keep}
    with pytest.raises(ValueError, match="The remaining columns are not a subset of the columns in ``pattern_dict``."):
        _ = srs._try_drop_empty_column(column_name_to_drop, data_lines, filtered_dict)


def test_tdec_no_data(expected_pattern_dict, column_name_to_drop):
    """
    No data associated with header
    """
    data_lines = ['NMBR  LOCATION  LO  COMMENT', 'NONE']
    out = data_lines.copy()
    out[0] = ' '.join([col.title() for col in out[0].split()][:-1])
    assert out == srs._try_drop_empty_column(column_name_to_drop, data_lines, expected_pattern_dict)


def test_tdec_col_not_empty(expected_pattern_dict, column_name_to_drop):
    """
    Column not empty
    """
    data_lines = [
        'NMBR  LOCATION  LO  COMMENT',
        '8000  S14W96   232',
        '3020 N20E20  210 additional_info',
    ]
    with pytest.raises(ValueError, match="not all rows have the same number of values as the remaining columns."):
        _ = srs._try_drop_empty_column(column_name_to_drop, data_lines, expected_pattern_dict)


def test_tdec_col_not_match_pattern(expected_pattern_dict, column_name_to_drop):
    """
    `LOCATION` column doesn't match the correct pattern
    """
    data_lines = [
        'NMBR  LOCATION  LO  COMMENT',
        '8000  S14W926   232',
    ]
    with pytest.raises(ValueError, match="not all rows match the provided pattern."):
        _ = srs._try_drop_empty_column(column_name_to_drop, data_lines, expected_pattern_dict)


def test_tdec_colname_not_exist(data_lines, expected_pattern_dict):
    """
    Try remove column name that does not exist in data
    """
    with pytest.raises(ValueError, match="The column 'nonexistent_column_name' does not exist."):
        _ = srs._try_drop_empty_column('nonexistent_column_name', data_lines, expected_pattern_dict)



from pathlib import Path
from datetime import datetime
from urllib.request import urlopen

import astropy.units as u
from astropy.time import TimeDelta

from sunpy.util.exceptions import warn_user
from sunpy.util.net import get_filename
from sunpy.util.util import hash_file

__all__ = ['Cache']


class Cache:
    """
    Cache provides a way to download and cache files.

    Parameters
    ----------
    downloader: Implementation of `~sunpy.data.data_manager.downloader.DownloaderBase`
        Downloader object for downloading remote files.
    storage: Implementation of `~sunpy.data.data_manager.storage.StorageProviderBase`
        Storage to store metadata about the files.
    cache_dir: `str` or `pathlib.Path`
        Directory where the downloaded files will be stored.
    expiry: `astropy.units.quantity.Quantity` or `None`, optional
        The interval after which the cache is invalidated. If the expiry is `None`,
        then the expiry is not checked. Defaults to 10 days.
    """

    def __init__(self, downloader, storage, cache_dir, expiry=10*u.day):
        self._downloader = downloader
        self._storage = storage
        self._cache_dir = Path(cache_dir)
        self._expiry = expiry if expiry is None else TimeDelta(expiry)

    def download(self, urls, namespace='', redownload=False):
        """
        Downloads the files from the urls.

        Parameters
        ----------
        urls : `list` of path-like or one path-like
            A list of urls or a single url.
            The list is for urls of the same file but from different sources.
        namespace : `str`, optional
            A namespace to be used for the file name.
            Defaults to an empty string.
        redownload : `bool`, optional
            Whether to skip cache and redownload.
            Defaults to `False`.

        Returns
        -------
        `pathlib.Path`
            Path to the downloaded file.
        """
        if isinstance(urls, str | Path):
            urls = [urls]
        # Logic plan
        # 1. Check if the file is present in cache by url
        # 2. If present and it has not expired nor redownload, return the file path
        # 3. If not present or present and (expired or redownload), download the file and update cache
        #   a. If there is an error from the above steps, we will return the file from the cache if present
        for url in urls:
            cache_details = self._get_by_url(url)
            if cache_details:
                break
        # If we have a cache hit and we do not want to redownload it and its still valid
        # We want to just return the file
        if cache_details and not redownload and not self._has_expired(cache_details):
            return Path(cache_details['file_path'])
        try:
            file_path, file_hash, url = self._download_and_hash(urls, namespace)
            # We explicitly check for redownload or expiry:
            # This isn't needed as it should be caught by the above if statement
            # but this is more explicit
            if cache_details and (redownload or self._has_expired(cache_details)):
                self._storage.delete_by_key('url', cache_details['url'])
            self._storage.store({
                'file_hash': file_hash,
                'file_path': str(file_path),
                'url': url,
                'time': datetime.now().isoformat(),
             })
            return file_path
        except Exception as e:
            if not cache_details:
                raise e
            exception_msg = f"{e!r} \n"
            warn_user(
                f"{exception_msg}Due to the above error, you will be working with a stale version of the file in the cache."
            )
            return Path(cache_details['file_path'])

    def _has_expired(self, details):
        """
        Whether the url corresponding to details in cache has expired or not.

        Parameters
        ----------
        details : `dict`
            Details detached from cache.

        Returns
        -------
        `bool`
            Whether the url has expired or not.
        """
        time = details.get("time", datetime.now().isoformat())
        time = datetime.fromisoformat(time)
        return self._expiry and datetime.now() - time > self._expiry

    def get_by_hash(self, sha_hash):
        """
        Returns the details which is matched by hash if present in cache.

        Parameters
        ----------
        sha_hash : `str`
            SHA-256 hash of the file.
        """
        return self._storage.find_by_key('file_hash', sha_hash)

    def _get_by_url(self, url):
        """
        Returns the details which is matched by url if present in cache.

        Parameters
        ----------
        url : `str`
            URL of the file.
        """
        return self._storage.find_by_key('url', url)

    def _download_and_hash(self, urls, namespace):
        """
        Downloads the file and returns the path, hash, and URL it used to download.

        Parameters
        ----------
        urls : `list` of `str`
            List of URLs.

        Returns
        -------
        `pathlib.Path`, `str`, `str`
            Path to the downloaded file, SHA-256 hash, and the URL used.
        """
        errors = []
        for url in urls:
            try:
                path = self._cache_dir / (namespace + get_filename(urlopen(url), url))

                self._downloader.download(url, path, overwrite=True)
                shahash = hash_file(path)
                return path, shahash, url
            except Exception as e:
                errors.append(e)
        else:
            if len(errors) == 1:
                raise errors[0]
            msg = "Download failed for all URLs, the first error is shown above."
            raise RuntimeError(msg) from errors[0]



"""
Test cases for SOHO EITMap subclass
"""
import pytest

import astropy.units as u

import sunpy.map
from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.mapbase import SpatialPair
from sunpy.map.sources.soho import EITMap
from .helpers import _test_private_date_setters


@pytest.fixture()
def eit_map():
    """Creates an EITMap from a FITS file."""
    return sunpy.map.Map(get_test_filepath("EIT/efz20040301.000010_s.fits"))


def test_fitstoEIT(eit_map):
    """Tests the creation of EITMap using FITS."""
    assert isinstance(eit_map, EITMap)


def test_eitmap_coordinate_system(eit_map):
    assert eit_map.coordinate_system ==  SpatialPair(axis1='HPLN-TAN', axis2='HPLT-TAN')


def test_reference_date(eit_map):
    assert eit_map.reference_date.isot == "2004-03-01T00:00:10.515"


def test_date(eit_map):
    assert eit_map.date.isot == "2004-03-01T00:00:10.515"


def test_private_date_setters(eit_map):
    _test_private_date_setters(eit_map)


def test_is_datasource_for(eit_map):
    """Test the is_datasource_for method of EITMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert eit_map.is_datasource_for(eit_map.data, eit_map.meta)


def test_observatory(eit_map):
    """Tests the observatory property of the EITMap object."""
    assert eit_map.observatory == "SOHO"


def test_measurement(eit_map):
    """Tests the measurement property of the EITMap object."""
    assert eit_map.measurement.value in [195, 171]


def test_rsun(eit_map):
    """Tests the measurement property of the EITMap object."""
    assert u.allclose(eit_map.rsun_obs, 979.0701*u.arcsec)


def test_norm_clip(eit_map):
    # Tests that the default normalizer has clipping disabled
    assert not eit_map.plot_settings['norm'].clip


def test_wcs(eit_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    eit_map.pixel_to_world(0*u.pix, 0*u.pix)


def test_old_eit_date():
    eit_map = get_dummy_map_from_header(get_test_filepath("seit_00171_fd_19961211_1900.header"))
    assert eit_map.date.value == '1996-12-11T19:00:14.254'
    assert eit_map.reference_date.value == '1996-12-11T19:00:14.254'



"""
===============================
Plotting a Map without any Axes
===============================

This examples shows you how to plot a Map without any annotations at all, i.e.,
to save as an image.
"""
import matplotlib.pyplot as plt
import numpy as np

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE

##############################################################################
# Create a sunpy Map from the sample data.

smap = sunpy.map.Map(AIA_171_IMAGE)

##############################################################################
# Plot the Map without a frame.
# We can setup a frameless figure and an axes which spans the whole canvas.

figure = plt.figure(frameon=False)
ax = plt.axes([0, 0, 1, 1])
# Disable the axis
ax.set_axis_off()

# Plot the map.
# Since we are not interested in the exact map coordinates,
# we can simply use :meth:`~matplotlib.Axes.imshow`.
norm = smap.plot_settings['norm']
norm.vmin, norm.vmax = np.percentile(smap.data, [1, 99.9])
ax.imshow(smap.data,
          norm=norm,
          cmap=smap.plot_settings['cmap'],
          origin="lower")

# sphinx_gallery_defer_figures

##############################################################################
# At this point you could save the figure with :func:`~matplotlib.pyplot.savefig`
# or show it:

plt.show()



"""
========================================
Getting the observer location from a Map
========================================

How to access the observer location from a `~sunpy.map.Map` and interpret it.
"""
import matplotlib.pyplot as plt
import numpy as np

from astropy.constants import R_earth

import sunpy.map
from sunpy.coordinates import get_body_heliographic_stonyhurst
from sunpy.data.sample import AIA_171_IMAGE

###############################################################################
# We use the sunpy sample data.

aiamap = sunpy.map.Map(AIA_171_IMAGE)

###############################################################################
# You can access the observer coordinate with:

print(aiamap.observer_coordinate)

###############################################################################
# This provides the location of the SDO as defined in the header and is
# necessary to fully define the helioprojective coordinate system which
# depends on where the observer is. Let's see where this is with respect
# to Earth. SDO is a geosynchronous orbit with a semi-major axis of
# 42,164.71 km and an inclination of 28.05 deg.
# We will convert it to Geocentric Celestial Reference System (GCRS)
# whose center is at the Earth's center-of-mass.

sdo_gcrs = aiamap.observer_coordinate.gcrs
sun = get_body_heliographic_stonyhurst('sun', aiamap.date)

##############################################################################
# Let's plot the results. The green circle represents the Earth.
# This looks like the Earth is in the way of SDO's
# field of view but remember that it is also above the plane of this plot
# by its declination.

fig = plt.figure()
ax = fig.add_subplot(projection='polar')
circle = plt.Circle((0.0, 0.0), 1.0, transform=ax.transProjectionAffine + ax.transAxes, color="green",
                    alpha=0.4, label="Earth")
ax.add_artist(circle)
ax.text(0.48, 0.5, "Earth", transform=ax.transAxes)
ax.plot(sdo_gcrs.ra.to('rad'), sdo_gcrs.distance / R_earth, 'o', label=f'SDO {sdo_gcrs.dec:.2f}')
ax.plot(sun.lon.to('rad').value * np.ones(2), [1, 10], '-', label='to Sun', color='black')
ax.legend()

plt.show()



"""
This module provides a generic file reader for internal use.
"""
import re
import gzip
import pathlib

import fsspec

from sunpy.util.io import is_uri

try:
    from . import _fits as fits
except ImportError:
    fits = None

try:
    from . import _jp2
except ImportError:
    _jp2 = None

try:
    from . import ana
except ImportError:
    ana = None


__all__ = ['read_file', 'read_file_header', 'write_file', 'detect_filetype']

# File formats supported by sunpy.io
_KNOWN_EXTENSIONS = {
    ('fts', 'fits'): 'fits',
    ('jp2', 'j2k', 'jpc', 'jpt'): 'jp2',
    ('fz', 'f0'): 'ana'
}


class Readers(dict):
    # Define a dict which raises a custom error message if the value is None
    def __init__(self, *args):
        dict.__init__(self, *args)

    def __getitem__(self, key):
        val = dict.__getitem__(self, key)
        if val is None:
            raise ReaderError(
                f"The Reader sunpy.io.{key} is not available, "
                "please check that you have the required dependencies "
                "installed."
            )
        return val

# File readers supported by sunpy.io
_READERS = Readers({
    'fits': fits,
    'jp2': _jp2,
    'ana': ana
})


def _read(filepath, function_name, filetype=None, **kwargs):
    """
    This functions provides the logic paths for reading a file.

    It is trying to handle, reading files with known extensions and reading files with known readers.

    We also want the detection logic to first try to detect the filetype based on the content and then
    fallback to using extension.

    Parameters
    ----------
    filepath : pathlib.Path, str
        The file to be read.
    function_name : {'read' | 'get_header'}
        The name of the function to call on the reader.
    filetype : {'jp2' | 'fits' | 'ana'}, optional
        Supported reader or extension to manually specify the filetype.
        Supported readers are ('jp2', 'fits', 'ana')
    **kwargs : `dict`
        Additional keyword arguments are handed to file specific reader.

    Returns
    -------
    pairs : `list`
        A list of (data, header) tuples.
    """
    filepath = str(filepath)
    if filetype is not None:
        return getattr(_READERS[filetype], function_name)(filepath, **kwargs)
    try:
        readername = detect_filetype(filepath)
        if readername in _READERS.keys():
            return getattr(_READERS[readername], function_name)(filepath, **kwargs)
        readername = None
    except UnrecognizedFileTypeError:
        readername = None
    for extension, name in _KNOWN_EXTENSIONS.items():
        if filepath.endswith(extension) or filetype in extension:
            readername = name
    if readername is not None:
        return getattr(_READERS[readername], function_name)(filepath, **kwargs)
    raise UnrecognizedFileTypeError("The requested filetype is not currently supported by sunpy.")


def read_file(filepath, filetype=None, **kwargs):
    """
    Automatically determine the filetype and read the file.

    Parameters
    ----------
    filepath : `str`, path-like
        The file to be read.
    filetype : `str`, optional
        Supported reader or extension to manually specify the filetype.
        Supported readers are ('jp2', 'fits', 'ana')
    memmap : `bool`, optional
        Should memory mapping be used, i.e. keep data on disk rather than in RAM.
        This is currently only supported by the FITS reader.
    **kwargs : `dict`
        All extra keyword arguments are passed to ``.read`` for the file specific reader.

    Returns
    -------
    pairs : `list`
        A list of (data, header) tuples.
    """
    return _read(filepath, 'read', filetype, **kwargs)


def read_file_header(filepath, filetype=None, **kwargs):
    """
    Reads the header from a given file.

    This should always return a instance of `~sunpy.io._header.FileHeader`.

    Parameters
    ----------
    filepath : `str`
        The file from which the header is to be read.
    filetype : `str`
        Supported reader or extension to manually specify the filetype.
        Supported readers are ('jp2', 'fits').
    **kwargs : `dict`
        All extra keyword arguments are passed to ``.get_header`` for the file specific reader.

    Returns
    -------
    headers : `list`
        A list of headers.
    """
    return _read(filepath, 'get_header', filetype, **kwargs)


def write_file(fname, data, header, filetype='auto', **kwargs):
    """
    Write a file from a data & header pair using one of the defined file types.

    Parameters
    ----------
    fname : `str`
        Filename of file to save.
    data : `numpy.ndarray`
        Data to save to a fits file.
    header : `collections.OrderedDict`
        Meta data to save with the data.
    filetype : `str`, {'auto', 'fits', 'jp2'}, optional
        Filetype to save if ``auto`` the  filename extension will
        be detected, else specify a supported file extension.
    **kwargs : `dict`
        All extra keyword arguments are passed to ``.write`` for the file specific reader.

    Notes
    -----
    * This routine currently only supports saving a single HDU.
    """
    if filetype == 'auto':
        # Get the extension without the leading dot
        filetype = pathlib.Path(fname).suffix[1:]
    for extension, readername in _KNOWN_EXTENSIONS.items():
        if filetype in extension:
            return _READERS[readername].write(fname, data, header, **kwargs)
    raise ValueError(f"The filetype provided ({filetype}) is not supported")


def detect_filetype(filepath, **kwargs):
    """
    Attempts to determine the type of file a given filepath is.

    Parameters
    ----------
    filepath : `str`, `pathlib.Path`
        Where the file is.
    **kwargs : `dict`
        Additional keyword arguments.

    Returns
    -------
    filetype : `str`
        The type of file.
    """
    if str(filepath).startswith('http') or  str(filepath).startswith('ftp'):
        return None
    if is_uri(filepath):
        fsspec_kw = kwargs.get("fsspec_kwargs", {})
        try:
            fileobj = fsspec.open(filepath, 'rb', **fsspec_kw).open()
        except Exception:
            return None
    else:
        fileobj = open(filepath, 'rb')
    with fileobj as fp:
        line1 = fp.readline()
        line2 = fp.readline()
        # Some FITS files do not have line breaks at the end of header cards.
        fp.seek(0)
        first80 = fp.read(80)
        # First 8 bytes of netcdf4/hdf5 to determine filetype as have same sequence
        fp.seek(0)
        first_8bytes = fp.read(8)
        # First 4 bytes of CDF
        fp.seek(0)
        cdf_magic_number = fp.read(4).hex()

    # For ASDF files
    if first80.startswith(b"#ASDF"):
        return "asdf"

    # FITS
    # Checks for gzip signature.
    # If found, decompresses first few bytes and checks for FITS
    if first80[:3] == b"\x1f\x8b\x08":
        with gzip.open(filepath, 'rb') as fp:
            first80 = fp.read(80)

    # Check for "KEY_WORD  =" at beginning of file
    match = re.match(br"[A-Z0-9_]{0,8} *=", first80)
    if match is not None:
        return 'fits'

    # JPEG 2000
    # Checks for one of two signatures found at beginning of all JP2 files.
    # Adapted from ExifTool
    # [1] https://www.sno.phy.queensu.ca/~phil/exiftool/
    # [2] https://www.hlevkin.com/hlevkin/Standards/fcd15444-2.pdf
    # [3] https://www.hlevkin.com/hlevkin/Standards/fcd15444-1.pdf
    jp2_signatures = [b"\x00\x00\x00\x0cjP  \x0d\x0a\x87\x0a",
                      b"\x00\x00\x00\x0cjP\x1a\x1a\x0d\x0a\x87\x0a"]
    for sig in jp2_signatures:
        if line1 + line2 == sig:
            return 'jp2'

    # netcdf4 and hdf5 files
    if first_8bytes == b'\x89HDF\r\n\x1a\n':
        return 'hdf5'

    if cdf_magic_number in ['cdf30001', 'cdf26002', '0000ffff']:
        return 'cdf'

    raise UnrecognizedFileTypeError("The requested filetype is not currently supported by sunpy.")


class UnrecognizedFileTypeError(OSError):
    """
    Exception to raise when an unknown file type is encountered.
    """


class ReaderError(ImportError):
    """
    Exception to raise when a reader errors.
    """



"""
This module provides general utility functions.
"""
import os
import hashlib
import inspect
from io import BytesIO
from base64 import b64encode
from shutil import get_terminal_size
from itertools import chain, count
from collections import UserList
from collections.abc import Iterator

import numpy as np

from astropy.wcs.utils import pixel_to_pixel

__all__ = ['unique', 'replacement_filename', 'expand_list',
           'expand_list_generator', 'dict_keys_same', 'hash_file', 'get_width',
           'get_keywords', 'get_set_methods', 'fix_duplicate_notes',
           'grid_perimeter', 'extent_in_other_wcs']

def unique(itr, key=None):
    """
    Return only unique elements of a sequence.

    Parameters
    ----------
    itr : `iterable`
        Any iterable sequence.
    key : `function`, optional
        A function to apply to each element in the iterable. Defaults to `None`.

    Returns
    -------
    `set`:
        A `set` of each unique element.
    """
    items = set()
    if key is None:
        for elem in itr:
            if elem not in items:
                yield elem
                items.add(elem)
    else:
        for elem in itr:
            x = key(elem)
            if x not in items:
                yield elem
                items.add(x)


def replacement_filename(path: str):
    """
    Return a replacement path if input path is currently in use.

    Enumerates until an unused filename is found, e.g., "foo.fits" becomes
    "foo.0.fits", if that is used, "foo.1.fits" and so on.

    Parameters
    ----------
    path : `str`
        A string path.

    Returns
    -------
    `str`:
        A string path.
    """
    if not os.path.exists(path):
        return path
    else:
        dir_, filename = os.path.split(path)
        base, ext = os.path.splitext(filename)
        for c in count():
            name = base + '.' + str(c) + ext
            newpath = os.path.join(dir_, name)
            if not os.path.exists(newpath):
                return newpath


def expand_list(inp):
    """
    Expand a list of lists or tuples.

    Parameters
    ----------
    inp : `list`, `tuple`, `collections.UserList`
        The iterable to expand.

    Returns
    -------
    `list`
        A flat list consisting of the entries of the input.

    References
    ----------
    * https://stackoverflow.com/questions/2185822/expanding-elements-in-a-list/2185971#2185971
    """
    return list(expand_list_generator(inp))


def expand_list_generator(inp):
    for item in inp:
        # parfive.Results are UserList
        if isinstance(item, (list | tuple | UserList | Iterator)):
            yield from expand_list_generator(item)
        else:
            yield item


def partial_key_match(key, dictionary):
    """
    Return value/values from a dictionary based on a partial key.

    Each element of the partial key is matched against the keys of the dictionary and
    if a partial match is found the value of the key is returned.

    Even a partial match works here i.e even if the key matches partially a value is returned.

    Parameters
    ----------
    key : `tuple`
        A tuple containing the partial key.
    dictionary : `dict`
        The target dictionary from which we want to retrieve the value based on the partial key.

    Yields
    ------
    value
        The value of the matched key.

    References
    ----------
    * https://stackoverflow.com/questions/18893624/partial-match-dictionary-keyof-tuples-in-python

    Examples
    --------
    >>> d = {('abc','def','ghi') : 1, ('abc', 'def', 'xyz') : 2, ('pqr', 'lmn', 'tuv') : 3}
    >>> list(partial_key_match(('abc', 'def', None), d))
        [1, 2]
    """
    for k, v in dictionary.items():
        if all(k1 == k2 or k2 is None for k1, k2 in zip(k, key)):
            yield v


def dict_keys_same(list_of_dicts):
    """
    Makes sure that a list of dictionaries all have the same keys.

    If a key is missing, it will be added but with a value of None.

    Parameters
    ----------
    list_of_dicts : `list` of `dict`
          A list containing each dictionary to parse.

    Returns
    ------
    `list`
        The list with each dict updated.

    References
    ----------
    * https://stackoverflow.com/questions/10482439/make-sure-all-dicts-in-a-list-have-the-same-keys

    Examples
    --------
    >>> l = [{'x': 42}, {'x': 23, 'y': 5}]
    >>> dict_keys_same(l)
        [{'x': 42, 'y': None}, {'x': 23, 'y': 5}]
    """
    all_keys = set(chain.from_iterable(list_of_dicts))
    for d in list_of_dicts:
        d.update({key: None for key in all_keys if key not in d})
    return list_of_dicts


def hash_file(path):
    """
    Returns the SHA-256 hash of a file.

    Parameters
    ----------
    path : `str`
        The path of the file to be hashed.

    Returns
    -------
    `str`
        SHA-256 hash of the file.

    References
    ----------
    * https://stackoverflow.com/a/22058673
    """
    BUF_SIZE = 65536
    sha256 = hashlib.sha256()

    with open(path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)

    return sha256.hexdigest()


def get_width():
    """
    Gets the width of the current terminal.
    Accounts for if the 'COLUMNS' environmental variable is set.

    Returns
    -------
    `int`
        Width of the terminal you are in.
        Works for IPython notebooks and normal terminals.
    """
    width = os.environ.get("COLUMNS", None)
    if width:
        width = int(width)
    else:
        width, _ = get_terminal_size()
    return width


def get_keywords(func):
    """
    Returns a set of keyword names from ``func``'s signature.
    Recursive if ``func`` is a list of functions and methods.

    Parameters
    ----------
    func : function or method or `list`
        Function or method (or list of those) to extract a
        set of accepted keyword arguments for.

    Returns
    -------
    keywords : `set`
        A set of accepted keyword arguments.
    """
    if isinstance(func, list):
        keywords = set()
        for f in func:
            keywords.update(get_keywords(f))
        return keywords
    sig = inspect.signature(func)
    keywords = {param.name for param in sig.parameters.values()
                if param.default is not inspect.Parameter.empty}
    return keywords


def get_set_methods(obj):
    """
    Returns a set of keyword names that can be handled by
    an object's ``set_...`` methods.

    Parameters
    ----------
    obj : `object`
        Matplotlib object such as `~matplotlib.image.AxesImage`
        to extract handled keyword arguments for.

    Returns
    -------
    keywords : `set`
        A set of accepted keyword arguments.

    Notes
    -----
    See :meth:`matplotlib.artist.Artist.update` for an example
    of Matplotlib relying on this capability.
    """
    return {
        m[4:] for m in dir(obj)
        if m.startswith('set_') and callable(getattr(obj, m, None))
    }


def _figure_to_base64(fig):
    # Converts a matplotlib Figure to a base64 UTF-8 string
    buf = BytesIO()
    fig.savefig(buf, format='png', facecolor='none')  # works better than transparent=True
    return b64encode(buf.getvalue()).decode('utf-8')


def fix_duplicate_notes(notes_to_add, docstring):
    """
    Merges a note section into a docstring that may have a notes section.

    Parameters
    ----------
    notes_to_add : str
        The notes that need to be added (starting with ``"Notes"``).
    docstring
        The original docstring.

    Returns
    -------
    str
        Updated docstring with a single notes section.
    """
    docstring = inspect.cleandoc(docstring)  # not necessary on Python 3.13+
    existing_notes_pos = docstring.find('Notes\n-----')
    new_notes_pos = notes_to_add.find('Notes\n-----')
    if new_notes_pos == -1:
        raise RuntimeError("The notes to add does not appear to be a 'Notes' section.")
    notes_to_add_data = notes_to_add[new_notes_pos + len('Notes\n-----'):].strip()

    # Insert the notes before either numpydoc section that can come after 'Notes', otherwise at end
    references_pos = docstring.find("References\n----------")
    examples_pos = docstring.find("Examples\n--------")
    index = min({len(docstring), references_pos, examples_pos} - {-1})

    pieces = [docstring[:index],
              "\n\n" if index == len(docstring) else "",
              "Notes\n-----\n" if existing_notes_pos == -1 else "",
              f"{notes_to_add_data}",
              f"\n\n{docstring[index:]}" if len(docstring) > index else ""]
    return "".join(pieces)


def grid_perimeter(nx, ny):
    """
    Return a sequence of (x, y) grid points for the perimeter of a grid.

    The sequence represents an open path starting at (0, 0); traversing clockwise
    through (0, ny), (nx, ny), and (nx, 0); and then returning to (0, 0)
    exclusive (i.e., stopping at (1, 0)).

    Parameters
    ----------
    nx : `int`
        The number of grid cells in the X direction
    ny : `int`
        The number of grid cells in the Y direction

    Returns
    -------
    `numpy.ndarray`
        A Nx2 array of (x, y) grid points, where N is ``2 * (nx + ny)``
    """
    edges = [[np.zeros(ny), np.arange(ny)],  # left edge
             [np.arange(nx), np.full(nx, ny)],  # top edge
             [np.full(ny, nx), np.arange(ny, 0, -1)],  # right edge
             [np.arange(nx, 0, -1), np.zeros(nx)]]  # bottom edge
    return np.hstack(edges).T


def extent_in_other_wcs(original_wcs, target_wcs, *, method, original_shape=None, integers=False):
    """
    Returns the pixel extent of one WCS in a different WCS.

    Parameters
    ----------
    original_wcs : `~astropy.wcs.WCS`
        The original WCS
    target_wcs : `~astropy.wcs.WCS`
        The target WCS
    method : `str`
        The method for determining the extent: 'all', 'edges', or 'corners'
    original_shape: 2-element tuple
        The array shape of the original WCS.
        This is optional if it is already defined in ``original_wcs``
    integers : `bool`
        If `True`, round the output appropriately to integer values.
        Defaults to `False`.

    Returns
    -------
    min_x, max_x, min_y, max_y : `float` or `int`
        The pixel extent
    """
    ny, nx = original_wcs.array_shape if original_shape is None else original_shape
    if method == 'all':
        pixels = np.indices((nx + 1, ny + 1)).reshape((2, -1)) - 0.5
    elif method == 'edges':
        pixels = grid_perimeter(nx, ny).T - 0.5
    elif method == 'corners':
        pixels = np.array([[0, 0, nx, nx], [0, ny, ny, 0]]) - 0.5
    else:
        raise ValueError("The allowed options for `method` are 'all', 'edges', or 'corners'.")

    xy = np.stack(pixel_to_pixel(original_wcs, target_wcs, *pixels))

    if not np.all(np.isfinite(xy)):
        if method == 'corners':
            raise RuntimeError("The extent could not be automatically determined from the corners. "
                               "Try specifying 'all' or 'edges'.")
        elif method == 'edges':
            raise RuntimeError("The extent could not be automatically determined from the edges. "
                               "Try specifying 'all'.")
        else:
            raise RuntimeError("The extent could not be automatically determined because all of "
                               "the coordinates in the original WCS transformed to NaNs.")

    min_xy = np.min(xy, axis=1)
    max_xy = np.max(xy, axis=1)
    if integers:
        min_xy = (np.floor(min_xy + 0.5)).astype(int)
        max_xy = (np.ceil(max_xy - 0.5)).astype(int)
    return [min_xy[0], max_xy[0], min_xy[1], max_xy[1]]
